//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/audio/audio2text.c,v $
// $Id: audio2text.c,v 1.1 2005/11/03 21:56:08 ah Exp $ 
//
// Description: cut a pice out of a wav file into a new wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: audio2text.c,v 1.1 2005/11/03 21:56:08 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string>
#include <cerrno>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/IO/Convert.h>
#include <AH/IO/AudioFile.h>
USING_NAMESPACE(AH)

// --- local constants ---

//static const char* cmdLineOptions = "hnpstvV";
static const char* cmdLineOptions = "hpstvV";
static const char* version = "0.1";

// --- local variables ---

static int verbose = 0;
//static bool normalize = false;
static bool printTime = false;
static bool printTocPos = false;
static bool printSamNum = false;

// --- local classes ---

// --- local functions ---

static void usage(bool help = false)
{
    cerr << "Usage: audio2text [-" << cmdLineOptions << "] <audiofile> <textfile>\n"
	 << "       print the data from an audio file (WAV, MP3, MP+, Ogg, Pac, ...) file into\n"
	 << "       a text file.\n"
	 << "  options:\n"
	 << "   -h   - print this help message\n"
//	 << "   -n   - normalize into float range [-1..1]\n"
	 << "   -t   - print time in seconds\n"
	 << "   -s   - print sample number\n"
	 << "   -p   - print toc position\n"
	 << "   -v   - verbose mode (increase debug level)\n"
	 << "   -V   - print version\n"
	 << "   <audiofile>  - file containing audio data, supported file types:\n"
	 << "                  .wav           : PCM-WAV\n"
	 << "                  .mp3           : MP3\n"
	 << "                  .mpc .mp+ .mpp : MPEG+\n"
	 << "                  .ogg           : OGG-Vorbis\n"
	 << "                  .pac           : LPAC\n"
	 << "                  and all types supported by libsndfile"
	 << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << "  audio2text a.wav a.txt\n"
	 << "  - print just the data in a.wav as text into a.txt\n"
	 << "  audio2text -s a.wav a.txt\n"
	 << "  - print sample number and data in a.wav as text into a.txt\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "audio2text Vs " << version << " " 
	 << "(c) 2005 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


static std::string samples_to_toc(uint32 samples)
{
    uint32 minutes = samples / (44100 * 4 * 60);
    samples %= (44100 * 4 * 60);
    uint32 seconds = samples / (44100 * 4);
    samples %= (44100 * 4);
    uint32 sectors = samples / 2352;

    char buf[32];
    (void)sprintf(buf, "%02d:%02d:%02d", minutes, seconds, sectors);
    return buf;
}


template <typename SampleType>
void processFile(
    const string& inputFileName,
    const string& outputFileName,
    IOAudioFileOpener& afo,
    int stripBits = 0
)
{
    // create a wav file object for read:
    if (3 < verbose) cerr << "open input file \"" << inputFileName
			  << "\":" << endl;

    IOAudioFile<SampleType> in(inputFileName, afo, verbose);

    // create a text file stream for output, print header:
    ofstream out(outputFileName.c_str());
    out << "# ASCII representation of " << inputFileName << "\n";
    out << "# Samplerate / Hz    : " << in.getSampleRate() << "\n";
    out << "# Bits per sample    : " << sizeof(SampleType) * 8 - stripBits << "\n";
    out << "# Number of channels : " << in.getNumChannels() << "\n";
    out << "# Play time / sec    : ";
    if (0 == in.getNumSamples()) out << "unknown (from pipe)\n";
    else                         out << in.getTime() << "\n";
    out << "#\n";
    if (printSamNum) out << "sample\t";
    if (printTime)   out << "time/sec\t";
    if (printTocPos) out << "TOC-position\t";
    for (int32 chan = 0; chan < in.getNumChannels(); ++chan)
	out << "ch-" << chan+1 << "\t";
    out << "\n";
    out << "#--------------------------------------------------------------\n";

    // read all samples (until EOF):
    std::vector<SampleType> frame;
    frame.reserve(in.getNumChannels());
    uint32 curFrame = 0;
    int ret = 0;
    int divider = 1 << stripBits;
    do
    {
	if (printSamNum) out << curFrame+1 << "\t";
	if (printTime) out << in.getTime(curFrame) << "\t";
	if (printTocPos) out << samples_to_toc(curFrame) << "\t";

	frame.clear();
	ret = in.readFrame(frame);
	if (!frame.empty())
	{
	    for (uint32 chan = 0; chan < frame.size(); ++chan)
	    {
		SampleType value = frame[chan];
		if (divider != 1) value /= divider;
//	        if (normalize)
//		    out << value / (double)in.getMaxAmplitude() << "\t";
//	        else 
		    out << value << "\t";
	    }
	    out << "\n";
	}
	++curFrame;
    } while (ret > 0); 
}


//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static void process
(
    const string& inputFileName,
    const string& outputFileName
)
{
    if (2 < verbose) cerr << "process:" << endl;

    // create a wav file object for read:
    if (3 < verbose) cerr << "open input file \"" << inputFileName
			  << "\":" << endl;
    IOAudioFileOpener afo;
    IOAudioFile<int16> in(inputFileName, afo, verbose);
    if (verbose  > 3) cerr << in.getInfo();

    int bits = in.getSampleBits();
    if (bits > 0)
    {
	if (bits > 32)      processFile<double>(inputFileName, outputFileName, afo);
	else if (bits > 16) processFile<int32>(inputFileName, outputFileName, afo, 32-bits);
	else                processFile<int16>(inputFileName, outputFileName, afo, 16-bits);
    }
    else if (bits < 0)
    {
	if (bits < -32) processFile<double>(inputFileName, outputFileName, afo);
	else            processFile<float>(inputFileName, outputFileName, afo);
    }
    else
    {
	processFile<double>(inputFileName, outputFileName, afo);
    }

} // end of process()


// --- global functions ---


int main(int argc, char* argv[])
{
    //------------------------------
    // handle command line options:
    //------------------------------
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
//	case 'n': normalize = true;     break;
	case 't': printTime = true;     break;
	case 'p': printTocPos = true;   break;
	case 's': printSamNum = true;   break;
	case 'v': verbose++;            break;
	case 'h': usage(true);          exit(0);
	case 'V': print_version();      exit(0);
	default : usage();              exit(1);
	}
    }

    // the remaining arguments are audio file and text file:
    if (optind >= argc) 
    {
	cerr << "Error: missing audio and text files" << endl;
	usage();
	exit(1);
    }

    // the next argument is the audio file:
    string inputFileName(argv[optind++]);

    // the next argument is the text file:
    if (optind >= argc) 
    {
	cerr << "Error: missing text file" << endl;
	usage();
	exit(1);
    }
    string outputFileName(argv[optind++]);

    // check for too many arguments:
    if (optind < argc) 
    {
	cerr << "Error: too many arguments (last valid is \"" 
	     << inputFileName << "\")" << endl;
	usage();
	exit (1);
    }

    //---------------------
    // process input file:
    //---------------------
    try { process(inputFileName, outputFileName); }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
