//---------------------------------------------------------------------------
// Copyright (c) 2000 Andreas Huennebeck
//
// $Source: /home/ah/CvsTree/daten/src/audio/recmain.c,v $
// $Id: recmain.c,v 1.1 2000/02/26 00:24:03 ah Exp $
//
// Description: 
//
//---------------------------------------------------------------------------

// --- forward declarations ---

// --- necessary includes ---

#include <unistd.h>
#include <iostream>

#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
#include <AH/Audio/LinuxKernelPcm.h>
#include <AH/Audio/LinuxKernelMix.h>
USING_NAMESPACE(AH)

// --- local data ---

static char* cmdLineOptions = "b:d:hl:o:qr:t:v";

static string filename("stdout");
static bool quiet = false;
static bool verbose = false;
static int debug = 0;
static int sample_rate = 44100;
static int sample_bits = 16;
static int sample_chans = 2;
static int buf_time = 100;

// --- functions ---


/***********************************************************************
 *
 * Local Function: usage
 *
 * Description: print info about usage of rec_audio, then exit
 *
 ***********************************************************************/
static void usage (const char* errmsg)
{
    if (errmsg) cerr << errmsg << endl;
    cerr << "rec_audio: program to record audio data\n"
	 << "  usage: rec_audio [" << cmdLineOptions << "] <file> \n"
	 << "  -o <file> : define output file name (default is "
	 << filename << ")\n"
	 << "  -l <time> : define duration of recording (default is "
	 << "infinite)\n"
	 << "  -b <bits> : define number of bits per sample (default is " 
	 << sample_bits << " bits)\n"
	 << "  -r <rate> : define sample rate in Hz (default is "
	 << (double)sample_rate/1000 << "kHz)\n"
	 << "  -c <num>  : define number of channels (default is "
	 << sample_chans << " channels)\n"
	 << "  -t <ms>   : use read buffer for <ms> time (default is "
	 << buf_time << " ms)\n"
	 << "  -q        : be absolutely quiet\n"
	 << "  -v        : verbose mode\n"
	 << "  -d <level>: set debug level\n"
	 << "  -h        : print this error message\n"
	 << endl;
    exit(errmsg ? 1 : 0);
} /* end of local function usage() *************************************/



int main(int argc, char* argv[])
{
    /*----------------------------------------------
     * set defaults (override static definitions): 
     *---------------------------------------------*/
    int time_length = -1;

    /*--------------------------------------
     * loop through command line options : 
     *-------------------------------------*/
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'b': sample_bits = atoi(optarg);		break;
	case 'c': sample_chans = atoi(optarg);		break;
	case 'd': debug = atoi(optarg);			break;
	case 'l': time_length = 1000 * atoi(optarg);	break;
	case 'o': filename = optarg;			break;
	case 'q': quiet = true; verbose = false;	break;
	case 'r': sample_rate = atoi(optarg);		break;
	case 't': buf_time = atoi(optarg);		break;
	case 'v': quiet = false; verbose = true;	break;

	case '?': /* fall through */
	case 'h': usage(0);				break;
	default : usage("Unknown option");		break;
	}
    }
    
    /*-------------------------------------
     * reject invalid remaining options : 
     *------------------------------------*/
    if (verbose && (argc > optind))
    {
	string msg("Warning: the following options are ignored:\n- ");
	for (; argc > optind; optind++)
	    msg += string(argv[optind]) + " ";
	cerr << msg << endl;
    }

    if (verbose)
    {
	cerr << "specified recording parameters:" << endl;
	cerr << "- Duration  : " << time_length/1000 << " sec" << endl;
	cerr <<	"- Sampling  : " 
	     << (double)sample_rate/1000 << " kHz/" 
	     << sample_bits << " Bits/" 
	     << sample_chans << " channels" << endl;
    }

    unsigned char* buf = 0;
    try
    {
	AudioLinuxKernelPcm pcm;

	// open device:
	pcm.openDev(SystemDevice::RO);

	// set desired sampling parameter:
	pcm.setSampleRate(sample_rate);
	pcm.setSampleBits(sample_bits);
	pcm.setChannels(sample_chans);

	// get default settings:
	if (verbose)
	{
	    cerr << "used recording parameters:" << endl;
	    cerr << "- Duration  : " << time_length/1000 << " sec" << endl;
	    cerr << "- Sampling  : " 
		 << (double)sample_rate/1000 << " kHz/" 
		 << sample_bits << " Bits/" 
		 << sample_chans << " channels" << endl;
	}

	// prepare buffer for record/play:
	unsigned int msecsize = sample_rate * sample_bits/8 * sample_chans;
	msecsize /= 1000;

	unsigned int bufsize = msecsize * buf_time;
	if (bufsize < 1024) bufsize = 1024;
	unsigned char* buf = new unsigned char [bufsize];
	if (!buf) throw ErrorSysOutOfMem();
	if (debug & 0x1) cerr << "using buffer of " << bufsize/1024 
			      << " kByte size" << endl;

	// record some time:
	if (!quiet)
	{ 
	    cout << "now recording " << time_length/1000 << " seconds: "; 
	    cout.flush();
	}
	pcm.sync();
	
	int rec_time = 0;
	bool finished = false;
	do
	{
	    if (debug & 0x4) cerr << "start: time_length=" << time_length 
				<< " ms, rec_time=" << rec_time 
				<< " ms" << endl; 

	    /* calculate current buffer size (reduce at end) : */
	    int curtime = buf_time;
	    unsigned int curbufsize = bufsize;
	    if ((time_length > 0) && 
		(time_length - rec_time) < buf_time)
	    {
		curtime = time_length - rec_time;
		curbufsize = msecsize * curtime;
	    }
	    if (debug & 0x4) cerr << "buf_time=" << buf_time 
				<< " ms, curtime=" << curtime 
				<< " ms, curbufsize = " 
				<< curbufsize << " bytes" << endl; 

	    /* read audio data into buffer : */
	    if (debug & 0x2) cerr << "read  " << curbufsize/1024 
				<< "kBytes" << endl;
	    pcm.readDev(buf, curbufsize);

	    if (debug & 0x8)
	    {
		if (sample_bits > 8)
		{
		    const short* dp = (const short*)buf;
		    for (unsigned int i = 0; i < 10; i++) 
			cout << *(dp+i) << " ";
		    cout << endl;
		}
		else
		{
		    for (unsigned int i = 0; i < 10; i++) 
			cout << (int)buf[i] << " ";
		    cout << endl;
		}
	    }
	    /* write buffer into file: */
	    if (debug & 0x2) cerr << "write " << curbufsize/1024 
				<< "kBytes" << endl;

	    rec_time += curtime;
	    if ((time_length > 0) && (rec_time >= time_length))
		finished = true;
	}
	while(!finished);

	if (!quiet) cout << "done." << endl;


	delete[] buf;
	// close device:
	pcm.closeDev();
    }
    catch(ErrorBasic& err)
    {
	if (buf) delete[] buf;
	cerr << "Error: " << err.getMsg() << endl;
	return 1;
    }
    catch(...)
    {
	if (buf) delete[] buf;
	cerr << "Error: unknown exception" << endl;
	return 1;
    }
    return 0;
}

// *********************************************************************
// *			     E N D   O F   F I L E                     *
// *********************************************************************
