//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/ft/RedSpec.c,v $
// $Id: RedSpec.c,v 1.7 2004/04/21 20:40:35 ah Exp $ 
//
// Description: print information about specified wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: RedSpec.c,v 1.7 2004/04/21 20:40:35 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include

#include <AH/Error/Intern.h>
#include <AH/Error/Sys.h>
#include <AH/Etech/Types.h>
#include <AH/Etech/Algo.h>
USING_NAMESPACE(AH)
#include "RedSpec.h"

#include <cstdlib>
#include <cmath>
#include <iostream>

// --- local constants ---

// --- local definitions ---

// --- local classes ---

#ifdef HAS_NO_LOG2
static inline double log2(double val)
{
    return log(val) / log((double)2);
}
#endif


ReducedSpectrum::ReducedSpectrum(unsigned int in_size, unsigned int out_size,
				 double max_freq, unsigned int debug)
    : inum(in_size), onum(out_size), 
      f_max(max_freq), f_inc(f_max / (double)(inum-1)),
      debon(debug),
      mean_abs_spectrum(0), mean_phi_spectrum(0),
      max_abs_spectrum(0), max_phi_spectrum(0),
      min_abs_spectrum(0), min_phi_spectrum(0),
      frequencies(0)
{
    if (onum < 2) throw ErrorInternBadArg("ReducedSpectrum: onum");
    if (inum < 2) throw ErrorInternBadArg("ReducedSpectrum: inum");
    if (onum > inum) throw ErrorInternBadArg("ReducedSpectrum: onum > inum");
}



ReducedSpectrum::~ReducedSpectrum()
{
    if (onum != inum)
    {
	delete[] min_phi_spectrum;
	delete[] min_abs_spectrum;
	delete[] max_phi_spectrum;
	delete[] max_abs_spectrum;
    }
    delete[] frequencies;
    delete[] mean_phi_spectrum;
    delete[] mean_abs_spectrum;
}



uint32 ReducedSpectrum::process(Cplx* spectrum, bool linear)
{
    mean_abs_spectrum = new double[onum+1];
    mean_phi_spectrum = new double[onum+1];
    frequencies = new double[onum+1];
    if (!mean_abs_spectrum || !mean_phi_spectrum || !frequencies)
	throw ErrorSysOutOfMem();

    if (onum == inum)
    {
	if (debon) std::cerr << "keep spectrum size as " << inum << std::endl;
	max_abs_spectrum = min_abs_spectrum = mean_abs_spectrum;
	max_phi_spectrum = min_phi_spectrum = mean_phi_spectrum;

	uint32 bad_cnt = 0;
	for (uint32 i = 0; i < inum ; ++i)
	{
	    frequencies[i] = f_inc * i;
	    double betrag = abs(spectrum[i]);
	    if (betrag > 0.0) max_abs_spectrum[i] = inVdB(betrag);
	    else
	    {
		max_abs_spectrum[i] = -1000;
		++bad_cnt;
	    }
	    max_phi_spectrum[i] = arg(spectrum[i]);
	    // handle invalid data
//	    if (i && (-1000 == max_abs_spectrum[i-1])) 
//		max_abs_spectrum[i-1] = max_abs_spectrum[i]; 
	}
	return bad_cnt;
    }

    // get memory:
    max_abs_spectrum = new double[onum+1];
    max_phi_spectrum = new double[onum+1];
    min_abs_spectrum = new double[onum+1];
    min_phi_spectrum = new double[onum+1];
    if (!max_abs_spectrum || !max_phi_spectrum ||
	!min_abs_spectrum || !min_phi_spectrum)
	throw ErrorSysOutOfMem();

    if (debon) std::cerr << "reduce spectrum size to " << onum << std::endl;
    double f_next = -1;
    double f1 = -1;
    double f2 = -1;
    double delta;
    uint32 ocnt = 0;
    uint32 sum_cnt = 0;

    double sum_abs = 0;
    double sum_phi = 0;
    double max_abs = 0;
    double max_phi = 0;
    double min_abs = 0;
    double min_phi = 0;

    if (linear)
    {
	delta = (f_max - f_inc) / (onum - 1);
	f_next = f_inc + delta;
    }
    else
    {
	delta = log2(f_max / f_inc) / (onum - 1);
	f_next = f_inc * pow(2.0, delta);
    }

    uint32 bad_cnt = 0;
    double f = 0;
    for (uint32 i = 0; i < inum ; ++i, f += f_inc)
    {
	if (0 == f) 
	{
	    // ignore DC point
	    frequencies[ocnt] = f;
	    double betrag = abs(spectrum[i]);
	    if (betrag > 0.0) mean_abs_spectrum[ocnt] = inVdB(betrag);
	    else
	    {
		mean_abs_spectrum[ocnt] = -1000;
		++bad_cnt;
	    }
	    mean_phi_spectrum[ocnt] = arg(spectrum[i]);
	    max_abs_spectrum[ocnt] = mean_abs_spectrum[ocnt];
	    max_phi_spectrum[ocnt] = mean_phi_spectrum[ocnt];
	    min_abs_spectrum[ocnt] = mean_abs_spectrum[ocnt];
	    min_phi_spectrum[ocnt] = mean_phi_spectrum[ocnt];
	    ++ocnt;
	    if (linear) f_next = f_inc + (ocnt * delta);
	    else        f_next = f_inc * pow(2.0, ocnt * delta);
	    sum_cnt = 0;
	    continue;
	}
	if (f > f_next)
	{
	    // reduce multiple points to one:
	    frequencies[ocnt] = sqrt(f1 * f2);
	    mean_abs_spectrum[ocnt] = sum_abs/(double)sum_cnt;
	    mean_phi_spectrum[ocnt] = sum_phi/(double)sum_cnt;
	    max_abs_spectrum[ocnt] = max_abs;
	    max_phi_spectrum[ocnt] = max_phi;
	    min_abs_spectrum[ocnt] = min_abs;
	    min_phi_spectrum[ocnt] = min_phi;
	    ++ocnt;
	    if (linear) f_next = f_inc + (ocnt * delta);
	    else        f_next = f_inc * pow(2.0, ocnt * delta);
	    sum_cnt = 0;
	    sum_abs = sum_phi = 0;
	}
	if (0 == sum_cnt)
	{
	    double betrag = abs(spectrum[i]);
	    if (betrag > 0.0) sum_abs = max_abs = min_abs = inVdB(betrag);
	    else
	    {
		sum_abs = max_abs = min_abs = -1000;
		++bad_cnt;
	    }
	    sum_phi = max_phi = min_phi = arg(spectrum[i]);
	    f1 = f2 = f;
	    ++sum_cnt;
	}
	else
	{
	    double betrag = abs(spectrum[i]);
	    double abs_val;
	    if (betrag > 0.0) abs_val = inVdB(betrag);
	    else
	    {
		abs_val = -1000;
		++bad_cnt;
	    }
	    double phi_val = arg(spectrum[i]);
	    sum_abs += abs_val;
	    sum_phi += phi_val;
	    if (abs_val > max_abs) max_abs = abs_val;
	    if (phi_val > max_phi) max_phi = phi_val;
	    if (abs_val < min_abs) min_abs = abs_val;
	    if (phi_val < min_phi) min_phi = phi_val;
	    f2 = f;
	    ++sum_cnt;
	}
    }
    if (sum_cnt && (ocnt < onum))
    {
	frequencies[ocnt] = sqrt(f1 * f2);
	mean_abs_spectrum[ocnt] = sum_abs/(double)sum_cnt;
	mean_phi_spectrum[ocnt] = sum_phi/(double)sum_cnt;
	max_abs_spectrum[ocnt] = max_abs;
	max_phi_spectrum[ocnt] = max_phi;
	min_abs_spectrum[ocnt] = min_abs;
	min_phi_spectrum[ocnt] = min_phi;
    }
    return bad_cnt;
}


// --- end of file ---




