//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/gensig/Signals.c,v $
// $Id: Signals.c,v 1.18 2005/10/29 10:44:04 ah Exp $ 
//
// Description: signal factory
//
// How to add a new signal
// -----------------------
// 1. include the corresponding header file "Sig<type>.h"
// 2. implement Signals::get<type>() member function
// 3. call Signals::get<type>() in Signals::add()
// 4. add usage string in Signals::help()
//
//---------------------------------------------------------------------------

const char sweep_CVSID[] = "$Id: Signals.c,v 1.18 2005/10/29 10:44:04 ah Exp $ (c) 2002 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)
#include "Signals.h"

#include <math.h>
#include <string.h>
USING_NAMESPACE(std)

#include <AH/Error/Sys.h>
#include <AH/Error/Intern.h>
#include <AH/STL/String.h>
USING_NAMESPACE(AH)

#include "SigDc.h"
#include "SigDirac.h"
#include "SigSinus.h"
#include "SigSquare.h"
#include "SigSaw.h"
#include "SigTriangle.h"
#include "SigSweep.h"
#include "SigRiaaSweep.h"
#include "SigRndSweep.h"
#include "SigPinkNoise.h"
#include "SigPinkNoiseRange.h"
#include "SigWhiteNoise.h"
#include "SigWhiteNoiseRange.h"
#include "SigMsb.h"
#include "SigLsb.h"
#include "SigMls.h"
#include "SigSilence.h"
#include "Util.h"

// --- local constants ---

// --- local data ---

// --- local classes ---

// --- local functions ---

// --- global functions ---


Signals::Signals
(
    unsigned int samplerate, unsigned int samplebits, double tim,
    unsigned int debug
) 
    : sr(samplerate), sb(samplebits), maxAmp(Util::getMaxAmplitude(sb)),
      t(tim), debon(debug) 
{
}



Signals::~Signals()
{
    for (unsigned int i = 0; i < signals.size(); ++i)
	delete signals[i];
}



void Signals::calculate(double* data, unsigned int num)
{
    for (unsigned int i = 0; i < signals.size(); ++i)
    {
	if (debon) cerr << "Signals::calculate: add signal #" << i+1 << endl;
	signals[i]->addSignal(data, num);
    }
}



void Signals::add(const std::string& argument)
{
    //------------------------------------------
    // parse the argument: cut at the komma sign
    //------------------------------------------
    std::vector<std::string> token;
    STLStringToken(argument, token, ",");
    for (unsigned int i = 0; i < token.size(); ++i)
	STLStringDelSurround(token[i]);
    if (0 == token.size()) throw ErrorBasic(argument + ": illegal argument");

    // parse the signal type:
    std::string type = token[0];
    if (type == "dc")
	signals.push_back(getDc(token));
    else if (type == "dirac")
	signals.push_back(getDirac(token));
    else if (type == "sinus")
	signals.push_back(getSinus(token));
    else if (type == "square")
	signals.push_back(getSquare(token));
    else if (type == "saw")
	signals.push_back(getSaw(token));
    else if (type == "triangle")
	signals.push_back(getTriangle(token));
    else if (type == "sweep")
	signals.push_back(getSweep(token));
    else if (type == "riaa sweep")
	signals.push_back(getRiaaSweep(token));
    else if (type == "random sweep")
	signals.push_back(getRandomSweep(token));
    else if (type == "white noise")
	signals.push_back(getWhiteNoise(token));
    else if (type == "white noise range")
	signals.push_back(getWhiteNoiseRange(token));
    else if (type == "pink noise")
	signals.push_back(getPinkNoise(token));
    else if (type == "pink noise range")
	signals.push_back(getPinkNoiseRange(token));
    else if (type == "msb")
	signals.push_back(getMsb(token));
    else if (type == "lsb")
	signals.push_back(getLsb(token));
    else if (type == "mls")
	signals.push_back(getMls(token));
    else if (type == "silence")
	signals.push_back(getSilence(token));
    else 
	throw ErrorBasic(type + ": unknown signal type");
}



std::string Signals::help(const std::string& header)
{
    std::string msg = "";

    msg += header + "\"sinus, <freq>, <amplitude>\"\n";
    msg += header + 
	"\"sweep, <freq-begin>, <freq-end>, lin/log, <amplitude>\"\n";
    msg += header + 
	"\"riaa sweep, <freq-begin>, <freq-end>, lin/log, <amplitude>\"\n";
    msg += header + 
	"\"random sweep, <freq-begin>, <freq-end>, lin/log, <amplitude>\"\n";
    msg += header + "\"square, <freq>, <amplitude>\"\n";
    msg += header + "\"saw, <freq>, <amplitude>\"\n";
    msg += header + "\"triangle, <freq>, <amplitude>\"\n";
    msg += header + "\"dirac, <freq>, <amplitude>\"\n";
    msg += header + "\"white noise, <amplitude>\"\n";
    msg += header + 
	"\"white noise range, <freq-begin>, <freq-end>, <amplitude>\"\n";
    msg += header + "\"pink noise, <amplitude>\"\n";
    msg += header + 
	"\"pink noise range, <freq-begin>, <freq-end>, <amplitude>\"\n";
    msg += header + "\"dc, <amplitude>\"\n";
    msg += header + "\"silence\"\n";
    msg += header + "\"msb, <freq>\"\n";
    msg += header + "\"lsb, <freq>, <bitnum (> 0)>\"\n";
    msg += header + "\"mls, <tap/tap/../tap>, <amplitude>\"\n";
    msg += header + "<freq> is a float number in Hz\n";
    msg += header + "<amplitude> is a float number, or in dB\n";
    return msg;
}



std::string Signals::mlsTable(const std::string& header)
{
    std::string msg = "";

    msg += header + "Table of MLS types\n";
    msg += header + "order   length  taps\n";
    msg += header + "---------------------------------\n";
    msg += header + "  2          3  2/1\n";
    msg += header + "  3          7  3/1\n";
    msg += header + "             7  3/2\n";
    msg += header + "  4         15  4/1\n";
    msg += header + "            15  4/3\n";
    msg += header + "  5         31  5/2\n";
    msg += header + "            31  5/3\n";
    msg += header + "  6         63  6/1\n";
    msg += header + "            63  6/5\n";
    msg += header + "  7        127  7/1\n";
    msg += header + "           127  7/6\n";
    msg += header + "  8        255  8/6/5/1\n";
    msg += header + "           255  8/5/3/2\n";
    msg += header + "  9        511  9/4\n";
    msg += header + "           511  9/5\n";
    msg += header + " 10       1023  10/3\n";
    msg += header + "          1023  10/7\n";
    msg += header + " 11       2047  11/2\n";
    msg += header + "          2047  11/9\n";
    msg += header + " 12       4095  12/7/4/3\n";
    msg += header + "          4095  12/11/8/6\n";
    msg += header + "          4095  12/11/10/2\n";
    msg += header + " 13       8191  13/4/3/1\n";
    msg += header + "          8191  13/12/10/9\n";
    msg += header + "          8191  13/12/11/9/6/5/2/1\n";
    msg += header + " 14      16383  14/12/11/1\n";
    msg += header + "         16383  14/13/8/4\n";
    msg += header + "         16383  14/13/12/2\n";
    msg += header + "         16383  14/12/10/9/7/5/3/1\n";
    msg += header + " 15      32767  15/1\n";
    msg += header + "         32767  15/14\n";
    msg += header + "         32767  15/11\n";
    msg += header + "         32767  15/8\n";
    msg += header + "         32767  15/12/11/8/7/6/4/2\n";
    msg += header + " 16      65535  16/5/3/2\n";
    msg += header + "         65535  16/15/13/4\n";
    msg += header + "         65535  16/12/11/10/7/4/3/2\n";
    msg += header + " 17     131071  17/3\n";
    msg += header + "        131071  17/14\n";
    msg += header + "        131071  17/14/13/9\n";
    msg += header + "        131071  17/14/11/9/6/5\n";
    msg += header + "        131071  17/15/13/11/10/9/8/4/2/1\n";
    msg += header + " 18     262143  18/7\n";
    msg += header + "        262143  18/11\n";
    msg += header + " 19     524287  19/6/5/1\n";
    msg += header + "        524287  19/18/17/14\n";
    msg += header + " 20    1048575  20/3\n";
    msg += header + "       1048575  20/17\n";
    msg += header + " 21    2097151  21/2\n";
    msg += header + "       2097151  21/19\n";
    msg += header + " 22    4194303  22/1\n";
    msg += header + "       4194303  22/21\n";
    msg += header + " 23    8388607  23/5\n";
    msg += header + "       8388607  23/18\n";
    msg += header + " 24   16777215  24/4/3/1\n";
    msg += header + "      16777215  24/23/22/17\n";
    msg += header + " 25   33554431  25/3\n";
    msg += header + " 26   67108863  26/8/7/1\n";
    msg += header + " 27  134217727  27/8/7/1\n";
    msg += header + " 28  268435455  28/3\n";
    msg += header + " 29  536870911  29/2\n";
    msg += header + " 30 1073741823  30/16/15/1\n";
    msg += header + " 31 2147483647  31/3\n";
    msg += header + " 32 4294967295  32/28/27/1\n";
    return msg;
}



SignalBase* Signals::getDc(const std::vector<std::string>& arg)
{
    // expected: "dc, ampl"
    if (2 > arg.size()) throw ErrorBasic("dc: not enough arguments");
    if (2 < arg.size()) throw ErrorBasic("dc: too many arguments");

    // parse the amplitude:
    std::string ampl = arg[1];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the dc signal object and add it to our list:
    SignalDc* sig = new SignalDc(a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getDirac(const std::vector<std::string>& arg)
{
    // expected: "dirac, freq, ampl"
    if (3 > arg.size()) throw ErrorBasic("dirac: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("dirac: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal dirac frequency");

    // parse the amplitude:
    std::string ampl = arg[2];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the dirac signal object and add it to our list:
    SignalDirac* sig = new SignalDirac(f, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getSinus(const std::vector<std::string>& arg)
{
    // expected: "sinus, f, ampl"
    if (3 > arg.size()) throw ErrorBasic("sinus: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("sinus: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal sinus frequency");

    // parse the amplitude:
    std::string ampl = arg[2];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the sinus signal object and add it to our list:
    SignalSinus* sig = new SignalSinus(f, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getSquare(const std::vector<std::string>& arg)
{
    // expected: "square, f, ampl"
    if (3 > arg.size()) throw ErrorBasic("square: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("square: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal square frequency");

    // parse the amplitude:
    std::string ampl = arg[2];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the square signal object and add it to our list:
    SignalSquare* sig = new SignalSquare(f, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getSaw(const std::vector<std::string>& arg)
{
    // expected: "saw, f, ampl"
    if (3 > arg.size()) throw ErrorBasic("saw: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("saw: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal saw frequency");

    // parse the amplitude:
    std::string ampl = arg[2];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the saw signal object and add it to our list:
    SignalSaw* sig = new SignalSaw(f, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getTriangle(const std::vector<std::string>& arg)
{
    // expected: "triangle, f, ampl"
    if (3 > arg.size()) throw ErrorBasic("triangle: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("triangle: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal triangle frequency");

    // parse the amplitude:
    std::string ampl = arg[2];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the triangle signal object and add it to our list:
    SignalTriangle* sig = new SignalTriangle(f, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getSweep(const std::vector<std::string>& arg)
{
    // expected: "sweep, f1, f2, lin/log, ampl"
    if (5 > arg.size()) throw ErrorBasic("sweep: not enough arguments");
    if (5 < arg.size()) throw ErrorBasic("sweep: too many arguments");

    // parse the start frequency:
    std::string freq = arg[1];
    double f1 = atof(freq.c_str());
    if (f1 <= 0) throw ErrorBasic(freq + ": illegal sweep start frequency");

    // parse the end frequency:
    freq = arg[2];
    double f2 = atof(freq.c_str());
    if (f2 <= 0) throw ErrorBasic(freq + ": illegal sweep end frequency");
    if (f2 == f1) throw ErrorBasic(
	freq + ": equal start and end frequency in sweep");

    // parse the mode:
    std::string mode = arg[3];
    bool lin;
    if (0 == strcasecmp("lin", mode.c_str())) lin = true;
    else if (0 == strcasecmp("log", mode.c_str())) lin = false;
    else throw ErrorBasic(mode + ": illegal sweep mode");
    if (!lin)
    {
	if (f1 == 0)
	    throw ErrorBasic(freq + ": illegal sweep start frequency");
	if (f2 == 0)
	    throw ErrorBasic(freq + ": illegal sweep end frequency");
    }

    // parse the amplitude:
    std::string ampl = arg[4];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the sweep signal object and add it to our list:
    SignalSweep* sig = new SignalSweep(f1, f2, lin, t, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getRiaaSweep(const std::vector<std::string>& arg)
{
    // expected: "riaa sweep, f1, f2, lin/log, ampl"
    if (5 > arg.size()) throw ErrorBasic("riaa sweep: not enough arguments");
    if (5 < arg.size()) throw ErrorBasic("riaa sweep: too many arguments");

    // parse the start frequency:
    std::string freq = arg[1];
    double f1 = atof(freq.c_str());
    if (f1 <= 0) throw ErrorBasic(freq + ": illegal sweep start frequency");

    // parse the end frequency:
    freq = arg[2];
    double f2 = atof(freq.c_str());
    if (f2 <= 0) throw ErrorBasic(freq + ": illegal sweep end frequency");
    if (f2 == f1) throw ErrorBasic(
	freq + ": equal start and end frequency in sweep");

    // parse the mode:
    std::string mode = arg[3];
    bool lin;
    if (0 == strcasecmp("lin", mode.c_str())) lin = true;
    else if (0 == strcasecmp("log", mode.c_str())) lin = false;
    else throw ErrorBasic(mode + ": illegal sweep mode");
    if (!lin)
    {
	if (f1 == 0)
	    throw ErrorBasic(freq + ": illegal sweep start frequency");
	if (f2 == 0)
	    throw ErrorBasic(freq + ": illegal sweep end frequency");
    }

    // parse the amplitude:
    std::string ampl = arg[4];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the RIAA sweep signal object and add it to our list:
    SignalRiaaSweep* sig = new SignalRiaaSweep(f1, f2, lin, t, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getRandomSweep(const std::vector<std::string>& arg)
{
    // expected: "random sweep, f1, f2, lin/log, ampl"
    if (5 > arg.size()) throw ErrorBasic("random sweep: not enough arguments");
    if (5 < arg.size()) throw ErrorBasic("random sweep: too many arguments");

    // parse the start frequency:
    std::string freq = arg[1];
    double f1 = atof(freq.c_str());
    if (f1 <= 0) throw ErrorBasic(freq + ": illegal sweep start frequency");

    // parse the end frequency:
    freq = arg[2];
    double f2 = atof(freq.c_str());
    if (f2 <= 0) throw ErrorBasic(freq + ": illegal sweep end frequency");
    if (f2 == f1) throw ErrorBasic(
	freq + ": equal start and end frequency in sweep");

    // parse the mode:
    std::string mode = arg[3];
    bool lin;
    if (0 == strcasecmp("lin", mode.c_str())) lin = true;
    else if (0 == strcasecmp("log", mode.c_str())) lin = false;
    else throw ErrorBasic(mode + ": illegal sweep mode");
    if (!lin)
    {
	if (f1 == 0)
	    throw ErrorBasic(freq + ": illegal sweep start frequency");
	if (f2 == 0)
	    throw ErrorBasic(freq + ": illegal sweep end frequency");
    }

    // parse the amplitude:
    std::string ampl = arg[4];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the random sweep signal object and add it to our list:
    SignalRandomSweep* sig = new SignalRandomSweep(f1, f2, lin, t, a, sr,
						   debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getWhiteNoise(const std::vector<std::string>& arg)
{
    // expected: "white noise, ampl"
    if (2 > arg.size()) throw ErrorBasic("white noise: not enough arguments");
    if (2 < arg.size()) throw ErrorBasic("white noise: too many arguments");

    // parse the amplitude:
    std::string ampl = arg[1];
    double a = Util::getAmplitude(ampl, maxAmp);

    // create the white noise signal object and add it to our list:
    SignalWhiteNoise* sig = new SignalWhiteNoise(a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getWhiteNoiseRange(const std::vector<std::string>& arg)
{
    // expected: "white noise range, f1, f2, ampl"
    if (4 > arg.size()) throw ErrorBasic("white noise: not enough arguments");
    if (4 < arg.size()) throw ErrorBasic("white noise: too many arguments");

    // parse the start frequency:
    std::string freq = arg[1];
    double f1 = atof(freq.c_str());
    if (f1 <= 0) throw ErrorBasic(freq + ": illegal range start frequency");

    // parse the end frequency:
    freq = arg[2];
    double f2 = atof(freq.c_str());
    if (f2 <= 0) throw ErrorBasic(freq + ": illegal range end frequency");
    if (f2 == f1) throw ErrorBasic(
	freq + ": equal start and end frequency in range");

    // parse the amplitude:
    std::string ampl = arg[3];
    double a = Util::getAmplitude(ampl, maxAmp);
    if (a == 0) throw ErrorBasic(ampl + ": illegal white noise amplitude");

    // create the white noise signal object and add it to our list:
    SignalWhiteNoiseRange* sig = new SignalWhiteNoiseRange(f1, f2, a,
							   sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getPinkNoise(const std::vector<std::string>& arg)
{
    // expected: "pink noise, ampl"
    if (2 > arg.size()) throw ErrorBasic("pink noise: not enough arguments");
    if (2 < arg.size()) throw ErrorBasic("pink noise: too many arguments");

    // parse the amplitude:
    std::string ampl = arg[1];
    double a = Util::getAmplitude(ampl, maxAmp);
    if (a == 0) throw ErrorBasic(ampl + ": illegal pink noise amplitude");

    // create the pink noise signal object and add it to our list:
    SignalPinkNoise* sig = new SignalPinkNoise(a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getPinkNoiseRange(const std::vector<std::string>& arg)
{
    // expected: "pink noise range, f1, f2, ampl"
    if (4 > arg.size()) throw ErrorBasic("pink noise: not enough arguments");
    if (4 < arg.size()) throw ErrorBasic("pink noise: too many arguments");

    // parse the start frequency:
    std::string freq = arg[1];
    double f1 = atof(freq.c_str());
    if (f1 <= 0) throw ErrorBasic(freq + ": illegal range start frequency");

    // parse the end frequency:
    freq = arg[2];
    double f2 = atof(freq.c_str());
    if (f2 <= 0) throw ErrorBasic(freq + ": illegal range end frequency");
    if (f2 == f1) throw ErrorBasic(
	freq + ": equal start and end frequency in range");

    // parse the amplitude:
    std::string ampl = arg[3];
    double a = Util::getAmplitude(ampl, maxAmp);
    if (a == 0) throw ErrorBasic(ampl + ": illegal pink noise amplitude");

    // create the pink noise signal object and add it to our list:
    SignalPinkNoiseRange* sig = new SignalPinkNoiseRange(f1, f2, a,
							   sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getMsb(const std::vector<std::string>& arg)
{
    // expected: "msb, freq"
    if (2 > arg.size()) throw ErrorBasic("msb: not enough arguments");
    if (2 < arg.size()) throw ErrorBasic("msb: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal msb frequency");

    // create the msb signal object and add it to our list:
    SignalMsb* sig = new SignalMsb(f, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getLsb(const std::vector<std::string>& arg)
{
    // expected: "lsb, freq, bits"
    if (3 > arg.size()) throw ErrorBasic("lsb: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("lsb: too many arguments");

    // parse the frequency:
    std::string freq = arg[1];
    double f = atof(freq.c_str());
    if (f <= 0) throw ErrorBasic(freq + ": illegal lsb frequency");

    // parse the bits:
    std::string bits = arg[2];
    int b = atoi(bits.c_str());
    if (b < 1) throw ErrorBasic(freq + ": illegal lsb bit number");

    // create the lsb signal object and add it to our list:
    SignalLsb* sig = new SignalLsb((unsigned int)b, f, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}


SignalBase* Signals::getMls(const std::vector<std::string>& arg)
{
    // expected: "mls, tap/tap/../tap, ampl"
    if (3 > arg.size()) throw ErrorBasic("mls: not enough arguments");
    if (3 < arg.size()) throw ErrorBasic("mls: too many arguments");

    // parse the bits:
    std::vector<std::string> taps;
    std::vector<unsigned int> tap;
    STLStringToken(arg[1], taps, "/");
    for (unsigned int i = 0; i < taps.size(); ++i)
    {
	STLStringDelSurround(taps[i]);
	tap.push_back(atoi(taps[i].c_str()));
    }
    if (0 == taps.size()) throw ErrorBasic(arg[2] + ": illegal argument");

    // parse the amplitude:
    std::string ampl = arg[2];
    double a = Util::getAmplitude(ampl, maxAmp);
    if (a == 0) throw ErrorBasic(ampl + ": illegal MLS amplitude");

    // create the mls signal object and add it to our list:
    SignalMls* sig = new SignalMls(tap, a, sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}



SignalBase* Signals::getSilence(const std::vector<std::string>& arg)
{
    // expected: "silence"
    if (1 > arg.size()) throw ErrorBasic("silence: not enough arguments");
    if (1 < arg.size()) throw ErrorBasic("silence: too many arguments");

    // create the dc signal object and add it to our list:
    SignalSilence* sig = new SignalSilence(sr, debon);
    if (!sig) throw ErrorSys("Signals::add", ENOMEM);
    return sig;
}


// --- end of file ---
