//---------------------------------------------------------------------------
// Copyright 2005 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// Additionally this source is licensed exclusively to the company
// Bruker BioSpin GmbH under the terms of the GNU Lesser General Public 
// License (LGPL). See the file 'COPYING.LIB' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/CUtil/SecureArray.h,v $
// $Id: SecureArray.h,v 1.3 2005/09/28 23:44:03 ah Exp $
//
//---------------------------------------------------------------------------

#ifndef	AH_CUTIL_SECURE_ARRAY_H		// prevent multiple includes
#define AH_CUTIL_SECURE_ARRAY_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
USING_NAMESPACE(AH)

#include <sys/types.h>

// --- types and classes ---

BEGIN_NAMESPACE(AH)

/** utilities for secure handling of C arrays in C++ code.

    Class CUtilSecureArray offers some functions to handle
    C style arrays (not pointers!) of arbitrary types in a
    secure way. 

    All functions are static and inline and can be used without
    object and without linking libCUtil.
 */
class CUtilSecureArray
{
 public:
    /** return number of array entries.
	@param array the array
	@return number of array entries
    */
    template<class T> 
	static size_t getNumEntries(const T& array)
	{
	    return sizeof(array)/sizeof(array[0]);
	}

    /** set all array entries to specified value.
	@param array the array
	@param val the value each array entry is set to
    */
    template<class Dest, class Source> 
	static void setVal(Dest& array, const Source& val)
	{
	    const size_t num = sizeof(array)/sizeof(array[0]);
	    for (size_t i = 0; i < num; ++i)
		array[i] = val;
	}

    /** copy all array entries from other array.

	copyAll() copies all array entries from 's' to 'd', but
	no more entries than 'd' can store. If 's' contains
	less entries than 'd' can store, then the remaining entries
	of 'd' are set to the default value 'defval'.

	copyAll() returns the difference of the entry numbers, that is 
	the number of 's' array entries minus the number of 'd' 
	array entries. So ideally copyAll() returns 0 if the arrays 
	have equal number of entries. If the return value is negative
	then the absolute of this number is the number of entries in
	'd' that have been set to 'defval'. If the return value is 
	positive then it is the number of entries in 's' that have not 
	been copied.

	@param d the destination array
	@param s the source array
	@param defval the default value
	@return number of 's' array entries minus number of 'd' array entries
    */
    template<class Dest, class Source, class Default>
	static int copyAll(Dest& d, const Source& s, const Default& defval)
	{
	    const int d_num = sizeof(d)/sizeof(d[0]);
	    const int s_num = sizeof(s)/sizeof(s[0]);
	    const int num = (s_num < d_num ? s_num : d_num);
	    for (int i = 0; i < num; ++i)
		d[i] = s[i];
	    for (int i = num; i < d_num; ++i)
		d[i] = defval;
	    return s_num - d_num;
	}

    /** copy all array entries from other array.

	copy() copies all array entries from 's' to 'd', but
	no more entries than 'd' can store. If 's' contains
	less entries than 'd' can store, then the remaining entries
	of 'd' are not set and keep their previous state.

	copy() returns the difference of the entry numbers, that is 
	the number of 's' array entries minus the number of 'd' 
	array entries. So ideally copy() returns 0 if the arrays 
	have equal number of entries. If the return value is negative
	then the absolute of this number is the number of entries in
	'd' that are not set. If the return value is positive then
	it is the number of entries in 's' that have not been copied.

	@param dest the destination array
	@param source the source array
	@return number of 's' array entries minus number of 'd' array entries
    */
    template<class Dest, class Source>
	static int copy(Dest& d, const Source& s)
	{
	    const int d_num = sizeof(d)/sizeof(d[0]);
	    const int s_num = sizeof(s)/sizeof(s[0]);
	    const int num = (s_num < d_num ? s_num : d_num);
	    for (int i = 0; i < num; ++i)
		d[i] = s[i];
	    return s_num - d_num;
	}
}; // class CUtilSecureArray


END_NAMESPACE

#endif // AH_CUTIL_SECURE_ARRAY_H

