//---------------------------------------------------------------------------
// Copyright 1999 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/IO/Riff/WaveFile.h,v $
// $Id: WaveFile.h,v 1.16 2004/04/21 20:50:26 ah Exp $
//
// Description: 
//
//---------------------------------------------------------------------------

#ifndef AH_IO_RIFF_WAVE_FILE_H		// prevent multiple includes
#define AH_IO_RIFF_WAVE_FILE_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
#include <AH/Config/Types.h>
USING_NAMESPACE(AH)

#include <cstdio>
#include <iostream>

// --- types and classes ---

BEGIN_NAMESPACE(AH)

// represents a RIFF WAVE file
class IORiffWaveFile
{
 public:
    //-------------------
    // con-/destructors:
    //-------------------
    
    /// constructor for input WAV file 
    IORiffWaveFile(const char* name, bool isPipe, int verbose = 0);

    /// constructor for input file (WAV, MP3, Ogg, ...) 
    IORiffWaveFile(const char* name, int verbose = 0);

    /// constructor for output WAV file 
    IORiffWaveFile
    (
	/// name of output file
	const char* name,
	/// number of channels
	uint16 channel_num,
	/// size of sample in bits
	uint16 sample_size,
	/// sample rate in Hz
	uint32 sample_rate,
	/// number of samples per channel
	uint32 sample_num,
	/// verbose level
	int verbose = 0
    );

    /// destructor
    virtual ~IORiffWaveFile();

    //------------------
    // access functions
    //------------------
    
    /// return true if wave file is a pipe
    bool isPipe() const { return waveIsPipe; }
    /// get number of samples
    uint32 getSampleNum() const { return waveSamples; }
    /// get number of channels
    uint16 getChannelNum() const { return channels; }
    /// get sample rate in Hz
    uint32 getSampleRate() const { return sampleRate; }
    /// get sample size in bits per sample
    uint16 getSampleSize() const { return sampleSize; }
    /// get maximum amplitude
    int32 getMaxAmplitude() const { return maxPosAmpl; }

    //-----------------------
    // read/write functions:
    //-----------------------

    /// read header of wave file
    void readHeader(const char* name, bool isPipe = false);
    /// read PCM samples from wave file
    /**
       <tt>readSamples()</tt> reads <tt>num</tt> samples into the
       <tt>data</tt> buffer. If the file contains more than one channel,
       then the samples are written in this order:
       <pre>
          sample 1 channel 1
          sample 1 channel 2
          sample 2 channel 1
          sample 2 channel 2
          sample 3 channel 1
          sample 3 channel 2
       </pre>
       and so on. Therefore the <tt>data</tt> buffer must have the size
       <tt>num * getChannelNum()</tt>. 
     */
    int32 readSamples(int32* data, uint32 num);
    /// read raw PCM data from wave file
    int32 readRaw(unsigned char* data, uint32 size);
    /// read a data buffer from wave file
    int32 readBuf(unsigned char* buf, uint32 size);

    /// rewrite header into new wave file
    void writeHeader(const char* name);
    /// write header into current wave file
    void writeHeader();
    // write PCM samples into wave file
    /**
       <tt>writeSamples()</tt> writes <tt>num</tt> samples from the
       <tt>data</tt> buffer. If the file contains more than one channel,
       then the samples are read from the <tt>data</tt> buffer in this
       order:
       <pre>
          sample 1 channel 1
          sample 1 channel 2
          sample 2 channel 1
          sample 2 channel 2
          sample 3 channel 1
          sample 3 channel 2
       </pre>
       and so on. Therefore the <tt>data</tt> buffer must have the size
       <tt>num * getChannelNum()</tt>. 
     */
    int32 writeSamples(const int32* data, uint32 num);
    /// write raw PCM data into new wave file
    int32 writeRaw(const unsigned char* data, uint32 size);
    // write a data buffer into wave file
    int32 writeBuf(const unsigned char* buf, uint32 size);

    //------------------
    // file positioning
    //------------------

    /// get current data position
    uint32 getDataPos() const;

    /// get data position from time/sec
    uint32 getDataPos(double t) const;

    /// get data position from specified sample number
    uint32 getDataPos(uint32 sampleNum) const;

    /// set file to sample position specified by time/sec
    uint32 setDataPos(double t);

    /// set file to specified sample number
    uint32 setDataPos(uint32 sampleNum);

    /// get current file position
    uint32 getFilePos() const;

    /// set file to specified file position
    uint32 setFilePos(uint32 pos);

    //----------------------
    // conversion utilities
    //----------------------

    /// get number of samples from time/sec
    uint32 getSampleNum(double t) const;

    /// get time/sec from number of samples
    double getTime(uint32 sampleNum) const;

    //-------------------
    // buffer utilities:
    //-------------------

    /// get size of data for desired time/sec
    uint32 getDataBufSize(double t) const;

    /// get size of data for desired samples
    uint32 getDataBufSize(uint32 sampleNum) const;

    /// get specified sample from buffer
    int32 getBufSample(unsigned char* buffer, uint32 sampleNum, int channel);

    /// set specified sample in buffer
    void setBufSample(unsigned char* buffer, uint32 sampleNum, int channel,
		      int32 sampleValue);

    //---------------------
    // analyzer functions:
    //---------------------

    /// analyze time frame
    void getLimits
    (
	/// starting point
	double	t1,
	/// end point
	double	t2,
	int16&	lMin,
	int16&	lMax,
	int16&	rMin,
	int16&	rMax,
	int32&	lMinCnt,
	int32&	lMaxCnt,
	int32&	rMinCnt,
	int32&	rMaxCnt
    );

    //-----------------
    // info functions:
    //-----------------

    /// print information about the wave file
    virtual std::ostream& printInfo(std::ostream& out);
    /// print data for specific time frame
    virtual std::ostream& printData
    (
	/// writable stream for data points
	std::ostream& out,
	/// number of data points
	size_t ptnum,
	/// starting point
	double t1,
	/// end point
	double t2
    );

 protected:
    const char*	waveName;
    int 	verbose;
 private:
    FILE*	waveFp;
    bool	waveIsPipe;
    uint32      pipePos;        // file position in case of pipe
    uint32	waveFileSize;
    uint32	waveDataSize;
    uint32	waveSamples;
    double	waveTime;
    uint16	channels;
    uint32	sampleRate;	// in samples per second
    uint16	sampleSize;	// in bits per sample
    int32	maxPosAmpl;	// maximum positive amplitude value
    int32	maxNegAmpl;	// maximum negative amplitude value
    uint32	fileStart;
    uint32	fileEnd;
    uint32	dataStart;
    uint32	dataEnd;
    uint32	curDataStart;
    uint32	curDataEnd;
};

END_NAMESPACE

#endif // !AH_IO_RIFF_WAVE_FILE_H
