//---------------------------------------------------------------------------
// Copyright 2005 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// Additionally this source is licensed exclusively to the company
// Bruker Analytik GmbH under the terms of the GNU Lesser General Public 
// License (LGPL). See the file 'COPYING.LIB' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/include/AH/Test/UnitTest.h,v $
// $Id: UnitTest.h,v 1.5 2006/04/06 22:21:58 ah Exp $
//
// Description: Test Logger Utility
//
//---------------------------------------------------------------------------

#ifndef AH_TEST_UNIT_TEST_H		// prevent multiple includes
#define AH_TEST_UNIT_TEST_H

// --- forward declarations ---

extern "C" int main(int argc, const char* argv[]);

// --- necessary includes ---

#include <AH/Config/Platform.h>
USING_NAMESPACE(AH)

#include <vector>
#include <string>
#include <iostream>
#include <sstream>
#include <exception>
#include <cmath>

USING_NAMESPACE(std)

// --- types and classes ---

BEGIN_NAMESPACE(AH)


/** the base class for unit tests.
    
  Class UnitTest is the base class for unit tests. It is used as follows:
  - You must derive a sub class and define the test functions which are called
    by the pure virtual function run(). 
  - Also run() has to be implemented.
 
 But both is quite simple, see the following example. Lets assume you have written
  a function int square(int) which returns the square of its argument, and
  you want to test it:

\verbatim 

#include "square.h"
#include <AH/Test/UnitTest.h>
using namespace AH;

// this is your test class:
class Test: public UnitTest
{
    // your test functions, to be called in Test::run():
    void test_valid_args();
    void test_invalid_args();

    // your function to run all your tests:
    virtual void run(const std::vector<std::string>&)
    {
        this->test_valid_args();
        this->test_invalid_args();
    }

 public:
    // your constructor - use cout for all messages
    Test(): UnitTest("function square()", std::cout) {}
}; 

// this is your single instance of your class Test 
UnitTest* theUnitTest =  new Test();

// your test function for testing valid arguments:
void Test::test_valid_args()
{
    this->announceTestSeries("test_valid_arguments");

    // run the tests: we expect no exceptions
    UNIT_TEST_BEGIN
    this->checkEqual("square(1)", square(1), 1); 
    this->checkEqual("square(2)", square(2), 4); 
    this->checkEqual("square(3)", square(3), 9); 
    this->checkEqual("square(-1)", square(-1), 1); 
    this->checkEqual("square(-2)", square(-2), 4); 
    this->checkEqual("square(-3)", square(-3), 9); 
    UNIT_TEST_END
}

// your test function for testing invalid arguments:
void Test::test_invalid_args()
{
    this->announceTestSeries("test_invalid_arguments");

    // run the test: we expect an exception of type SquareNullArg here
    UNIT_TEST_EXCEPTION_BEGIN(SquareNullArg)
    square(0);
    UNIT_TEST_EXCEPTION_END(SquareNullArg)
}

\endverbatim 
  Yes, this is all the code you have to write. No need to write
  a main() function, or to parse command line arguments, or to
    write try-catch blocks. All this is done by class UnitTest, the
    already implemented main() function and the macros. 

  The following command line options are already handled:
  - -q = quiet mode: only two lines message are written at all.
  - -v = verbose mode: print all necessary information for bug search. 
  - -d = increase debug level: print additional debug messages
  - -h = help: print help message.
 */
class UnitTest
{
 public:
    /** set verbose mode on or off - called by main(). */
    void setVerbose(bool onoff)
	{ this->verbose = onoff; if (this->verbose) this->quiet = false; }

    /** set quite mode on or off - called by main(). */
    void setQuiet(bool onoff)
	{ this->quiet = onoff; if (this->quiet) this->verbose = false; }

    /** set debug mode to specified level - called by main(). */
    void setDebugLevel(int level)
	{ this->debug = level; }

    /** print help message - called by main(). */
    void printHelp(const char* progName);

    /** run the tests. */
    virtual void run(const std::vector<std::string>& args) = 0;

    /** return true if any of the tests failed - called by main(). */
    bool testsFailed() const { return (0 != this->failedNum); }

    /** destructor. */
    virtual ~UnitTest();

    /** utility function: return any type as string.

        asString() returns any type as std::string if the operator<<()
	is implemented for this type.
	\param t the type to be returned as string
     */
    template<class T>
    static std::string asString(T t)
	{
	    std::ostringstream oss;
	    oss << t;
	    return oss.str();
	}

 protected:
    /** output stream for all messages. */
    std::ostream& out;

    /** constructor.
	\param name the name of the test target, e.g. "class foo" or "function bar".
	\param Out the output stream for all messages. Defaults to cout if omitted.
     */
    UnitTest(const std::string& name, std::ostream& Out = std::cout);

    /** return true if verbose mode is enabled. */
    bool isVerbose() const { return this->verbose; }

    /** return true if quiet mode is enabled. */
    bool isQuiet() const { return this->quiet; }

    /** return debug level. */
    int getDebugLevel() const { return this->debug; }

    /** print user defined help message. 

	printUserHelp() is called by printHelp() when the test program is
        called with option -h. Override it in your Test class if you want to
        add more help lines.
    */
    virtual void printUserHelp() {};

    /** print message followed by CR/LF (not in quiet mode though).
        \param msg the message to be printed.
     */
    void announceMessage(const std::string& msg)
	{ if (!this->isQuiet()) this->out << msg << ":" << std::endl; }

    /** announce new test series (not in quiet mode though).

	announceTestSeries() prints "--- test of 'series' ---" followed by CR/LF,
	but does nothing when in quiet mode.
        \param series the name of the test series
     */
    virtual void announceTestSeries(const std::string& series)
	{ if (!this->quiet) this->out << "------ Test of " << series << " ------" << std::endl; }

    /** announce next test (not in quiet mode though).

	announceNextTest() prints
	- "Test <num>: <target> :" in verbose mode
	- "Test <num>: " in normal mode
	- nothing in quiet mode.

	The test number is always incremented.

	You may call announceNextTest() before a call of a function which might
        throw an exception. In this case call printPassed() behind the function 
        call.
        \param target the name of the test target, e.g. "foo()"
     */
    virtual void announceNextTest(const std::string& target)
	{
	    ++this->testNum;
	    if (!this->isQuiet()) this->out << "Test " << this->testNum << ": ";
	    if (this->isVerbose() && target.length()) this->out << target << ": ";
	}

    /** print "failed" followed by CR/LF (not in quiet mode though).

	printFailed() prints "failed."  followed by CR/LF,
	but not when in quiet mode. The number of failures is incremented always.

	Usually this function need not be called by the user.
     */
    virtual void printFailed()
	{
	    ++this->failedNum;
	    if (!this->isQuiet()) this->out << "failed." << std::endl; 
	}

    /** print "got=<got>, expected=<expected>. failed." followed by CR/LF. 

	printFailed() prints
	- "got=<got>, expected=<expected>. failed." followed by CR/LF in verbose mode
	- "failed." followed by CR/LF in normal mode
	- nothing in quiet mode.

	The number of failures is incremented always.
	Usually this function need not be called by the user.
	\param got the result of a test
	\param expected the expected result of a test
     */
    template<class T1, class T2>
    void printFailed(const T1& got, const T2& expected)
	{ 
	    ++this->failedNum;
	    if (this->isVerbose()) this->out << "got=\"" << got << "\", expected=\"" << expected << "\": "; 
	    if (!this->isQuiet()) this->out << "failed." << std::endl; 
	}

    /** print "passed." followed by CR/LF. 

	printPassed() prints "passed." followed by CR/LF,
	but not when in quiet mode.

	You may call printPassed() behind a call of a 
	function which might throw an exception. In this case call announceNextTest()
	before the function call.
     */
    virtual void printPassed()
	{
	    if (!this->isQuiet()) this->out << "passed." << std::endl; 
	}

    /** print "got=<got>. passed." followed by CR/LF. 

	printPassed() prints
	- "got=<got>: passed." followed by CR/LF in verbose mode
	- "passed." followed by CR/LF in normal mode
	- nothing in quiet mode.

	Usually this function need not be called by the user.
	\param got the result of a test
     */
    template<class T>
    void printPassed(const T& got)
	{
	    if (this->isVerbose()) this->out << "got=\"" << got << "\": "; 
	    if (!this->isQuiet()) this->out << "passed." << std::endl; 
	}

    /** print "Test <num>: caught exception <exception>: failed. "followed by CR/LF. 

	printException() prints
	- "Test <num>: caught exception <exception.what()>: failed." followed by CR/LF in verbose mode
	- "Test <num>: failed." followed by CR/LF in normal mode
	- nothing in quiet mode.

	The number of failures is incremented always.
	This function need not be called by the user.
	\param ex any exception derived from std::exception
     */
    virtual void printException(const std::exception& ex)
	{
	    this->announceNextTest("");
	    ++this->failedNum;
	    if (this->isVerbose()) this->out << "caught exception \"" << ex.what() << "\": "; 
	    if (!this->isQuiet()) this->out << "failed." << std::endl; 
	}

    /** print "Test <num>: caught exception: <msg>: failed. "followed by CR/LF. 

	printException() prints
	- "Test <num>: caught exception <msg>: failed." followed by CR/LF in verbose mode
	- "Test <num>: failed." followed by CR/LF in normal mode
	- nothing in quiet mode.

	The number of failures is incremented always.
	This function need not be called by the user.
	\param msg the message of the exception
     */
    virtual void printException(const std::string& msg)
	{
	    this->announceNextTest("");
	    ++this->failedNum;
	    if (this->isVerbose()) this->out << "caught exception: \"" << msg << "\": "; 
	    if (!this->isQuiet()) this->out << "failed." << std::endl; 
	}

    /** print "Test <num>: caught exception: <msg>: failed. "followed by CR/LF.

	printException() prints
	- "Test <num>: caught exception <msg>: failed." followed by CR/LF in verbose mode
	- "Test <num>: failed." followed by CR/LF in normal mode
	- nothing in quiet mode.

	The number of failures is incremented always.
	This function need not be called by the user.
	\param msg the message of the exception
     */
    virtual void printException(const char* msg)
	{ this->printException(std::string(msg)); }

    /** print "missing exception <exception>: failed. "followed by CR/LF. 

	printExceptionFailed() prints
	- "missing exception <name>: failed." followed by CR/LF in verbose mode
	- "failed." followed by CR/LF in normal mode
	- nothing in quiet mode.

	The number of failures is incremented always.
	This function need not be called by the user.
	\param name the name of the exception
     */
    virtual void printExceptionFailed(const std::string& name)
	{
	    ++this->failedNum;
	    if (this->isVerbose()) this->out << "missing exception \"" << name << "\": "; 
	    if (!this->isQuiet()) this->out << "failed." << std::endl; 
	}

    /** check the assertion for true: print fail if not, otherwise passed.

	checkTrue() checks the assertion 'A'. If 'A' is true it
	- prints "got=1: passed." followed by CR/LF in verbose mode
	- prints "passed." followed by CR/LF in normal mode

	If 'A' is false it
	- prints "got=0: expected=1: failed." followed by CR/LF in verbose mode
	- prints "failed." followed by CR/LF in normal mode
	- increments the number of failures.

	In quiet mode nothing is printed.

	Call this function if you want to test a statement as true, but cannot
        call checkEqual().
	\param name the name of the test
	\param assertion the assertion to be tested for true
    */
    template<class A> 
    void checkTrue(const std::string& name, const A& assertion)
    {
	this->announceNextTest(name);
	if (!assertion)
	{
	    this->printFailed(false, true);
	}
	else
	{
	    this->printPassed(true);
	}
    }

    /** check the arguments for equality (==): print fail if not, otherwise passed

	checkEqual() checks whether 'got' is equal to 'expected'. If both are equal it
	- prints "Test <num>: <msg>: got=1: passed." followed by CR/LF in verbose mode
	- prints "Test <num>: passed." followed by CR/LF in normal mode

	If they are not equal it
	- prints "Test <num>: <msg>: got=<got>: expected=<expected>: failed." followed by CR/LF in verbose mode
	- prints "Test <num>: failed." followed by CR/LF in normal mode
	- increments the number of failures.

	In quiet mode nothing is printed.

	Call this function if you want to test two parameters for equality. You can
	do this in all cases where
	- the operator==() is implemented for both types, and
        - the operator<<() is implemented for both types. 

	If these operators are not implemented you must resort to checkTrue(). 
	\param name the name of the test
	\param got the result of a test
	\param expected the expected result of a test
     */
    template<class T1, class T2>
    void checkEqual(const std::string& name, const T1& got, const T2& expected)
    {
	this->announceNextTest(name);
	if (!(expected == got))
	{
	    this->printFailed(got, expected);
	}
	else
	{
	    this->printPassed(got);
	}
    }

    /** check the arguments for equality (==): print fail if not, otherwise passed.

        This is just a partly specialized version of checkEqual() for character pointers.
        Visual C++ seems to need this.
	\param msg the description of the test
	\param got the result of a test
	\param expected the expected result of a test
     */
    template<class T>
    void checkEqual(const std::string& msg, const char* got, const T& expected)
    {
	this->checkEqual(msg, std::string(got), expected);
    }

    /** check the arguments for equality (==): print fail if not, otherwise passed.

        This is just a partly specialized version of checkEqual() for character pointers.
        Visual C++ seems to need this.
	\param msg the description of the test
	\param got the result of a test
	\param expected the expected result of a test
     */
    template<class T>
    void checkEqual(const std::string& msg, const T& got, const char* expected)
    {
	this->checkEqual(msg, got, std::string(expected));
    }

    /** check the arguments for equality (==): print fail if not, otherwise passed.

        This is just a specialized version of checkEqual() for character pointers.
        Visual C++ seems to need this.
	\param msg the description of the test
	\param got the result of a test
	\param expected the expected result of a test
     */
    void checkEqual(const std::string& msg, const char* got, const char* expected)
    {
	this->checkEqual(msg, std::string(got), std::string(expected));
    }

    /** check the floating point arguments for equality (==) within epsilon range: print fail if not, otherwise passed.

        This is a specialized version of checkEqual() for comparison of floating point types 
        within a specified epsilon range.
	\param msg the description of the test
	\param got the result of a test
	\param expected the expected result of a test
	\param eps the size of the epsilon range
    */
    template<class T1, class T2>
    void checkEqual(const std::string& msg, const T1& got, const T2& expected, double eps)
    {
	this->announceNextTest(msg);
	if (fabs(expected - got) > eps)
	{
	    if (this->isVerbose()) this->out << "got=\"" << got << "\", expected=\"" << expected << "\"" << ": "; 
	    this->printFailed();
	}
	else
	{
	    this->printPassed(got);
	}
    }

    /** additional command line options. */
    std::string addOptions;

    /** additional command line arguments. */
    std::string addArguments;

 private:
    std::string testName;
    unsigned int testNum;
    unsigned int failedNum;
    int debug;
    bool verbose;
    bool quiet;
}; // class UnitTest

// do not use this macro directly
#define UNIT_TEST_STANDARD_CATCHES \
    catch(const std::exception& ex) { this->printException(ex); } \
    catch(...) { this->printException("unknown exception"); }

// do not use this macro directly
#define UNIT_TEST_PRINT_FILE_POSITION \
    if (this->getDebugLevel()) \
        this->out << "at " << __FILE__ << " line " << __LINE__ << std::endl;

/** macro: mark begin of normal test.

  Call this macro before normal tests where no exceptions are expected to be
  thrown. It opens a new block, so all instances and variables created below
  are destroyed with the macro UNIT_TEST_END.
 */
#define UNIT_TEST_BEGIN \
    try { \
        UNIT_TEST_PRINT_FILE_POSITION

/** macro: mark end of normal test. 

  Call this macro behind normal tests where no exceptions are expected to be
  thrown. It closes the block opened with the macro UNIT_TEST_BEGIN, and
  destroys all instances and variables created behind UNIT_TEST_BEGIN above.
 */
#define UNIT_TEST_END \
    } \
    UNIT_TEST_STANDARD_CATCHES


/** macro: mark begin of test for expected exception. 

  Call this macro before tests where a specified exception is expected to be
  thrown. It opens a new block, so all instances and variables created below
  are destroyed with the macro UNIT_TEST_EXCEPTION_END.
  \param ExpectedException the expected exception
*/
#define UNIT_TEST_EXCEPTION_BEGIN(ExpectedException) \
    try { \
	std::string unitTestExceptionName = #ExpectedException; \
        UNIT_TEST_PRINT_FILE_POSITION \
	this->announceNextTest("expect throw " + unitTestExceptionName);

/** macro: mark end of test for expected exception. 

  Call this macro behind tests where a specified exceptions is expected to be
  thrown. It closes the block opened with the macro UNIT_TEST_EXCEPTION_BEGIN, and
  destroys all instances and variables created behind UNIT_TEST_EXCEPTION_BEGIN above.
  \param ExpectedException the expected exception
 */
#define UNIT_TEST_EXCEPTION_END(ExpectedException) \
        this->printExceptionFailed(unitTestExceptionName); \
    } \
    catch(const ExpectedException&) { this->printPassed(#ExpectedException); } \
    UNIT_TEST_STANDARD_CATCHES

END_NAMESPACE

extern UnitTest* theUnitTest;

#endif // !AH_TEST_UNIT_TEST_H
