//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/miditools/stripSysex.c,v $
// $Id: stripSysex.c,v 1.1 2006/01/05 01:33:01 ah Exp $ 
//
// Description: strip leading and following 0xf8 bytes from specified midi file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: stripSysex.c,v 1.1 2006/01/05 01:33:01 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <vector>
#include <string>
#include <cerrno>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

// --- local constants ---

static const char* cmdLineOptions = "hvV";
static const char* version = "0.1";
static const int BUFSIZE = 1024;

typedef std::vector<unsigned char> MidiData;

// --- local functions ---


static void usage()
{
    cerr << "Usage: stripsysex [-" << cmdLineOptions << "] file [file, ...]\n"
	 << "       strip leading and following 0xf8 bytes from MIDI file(s).\n"
	 << "  options:\n"
	 << "    h   - print this help message\n"
	 << "    v   - increase verbose mode\n"
	 << "    V   - print version" << endl;
} // end of usage()


static void print_version()
{
    cerr << "stripsysex Vs " << version << " " 
	 << "(c) 2006 Andreas Huennebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andi@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


static void read_file(const std::string& filename, MidiData& bytes, int verbose)
{
    // open file for read:
    FILE* fp = fopen(filename.c_str(), "r");
    if (!fp)
    {
	if (verbose) std::cout << " [failed]" << std::endl;
	throw ErrorSys(filename);
    }

    // read all data into vector of bytes:
    bytes.clear();
    while (true)
    {
	unsigned char buf[BUFSIZE];
	size_t numRead = fread(buf, 1, sizeof(buf), fp);
	if (0 == numRead) break;
	if (0 > numRead) 
	{
	    (void)fclose(fp);
	    throw ErrorSys("Cannot read from " + filename);
	}
	bytes.insert(bytes.end(), buf, buf + numRead);
    }

    // close file:
    (void)fclose(fp);

} // end of read_file()


static void write_file(const std::string& filename, 
		       const MidiData& bytes, int verbose)
{
    // open file for write:
    FILE* fp = fopen(filename.c_str(), "w");
    if (!fp)
    {
	if (verbose) std::cout << " [failed]" << std::endl;
	throw ErrorSys(filename);
    }

    // write all data from vector of bytes:
    size_t numWrite = bytes.size();
    size_t bytePos = 0;
    while (numWrite)
    {
	// copy data from vector into buffer
	unsigned char buf[BUFSIZE];
	size_t toWrite = numWrite;
	if (toWrite > sizeof(buf)) toWrite = sizeof(buf);
	for (size_t i = 0; i < toWrite; ++i)
	    buf[i] = bytes.at(bytePos + i);

	// write buffer
	int written = fwrite(buf, 1, toWrite, fp);
	if (0 == written) break;
	if (0 > written) 
	{
	    (void)fclose(fp);
	    throw ErrorSys("Cannot write to " + filename);
	}
	bytePos += written;
	numWrite -= written;
    }

    // close file:
    (void)fclose(fp);
} // end of write_file()


static void strip_file(const std::string& filename, int verbose)
{
    if (verbose) 
    {
	std::cout << "strip file \"" << filename << "\" ...";
	std::cout.flush();
    }

    //-------------------------------------
    // read all data into vector of bytes:
    //-------------------------------------
    MidiData bytes;
    read_file(filename, bytes, verbose);

    //----------------------------------
    // strip data from vector of bytes:
    //----------------------------------
    const unsigned char STRIPVAL = 0xf8;

    // strip leading stuff;
    if (verbose > 1) std::cout << " stripped: ";
    MidiData::iterator start = bytes.begin();
    while ((start != bytes.end()) && (STRIPVAL == *start))
	++start;
    if (start != bytes.begin())
    {
	if (verbose > 1) std::cout << start - bytes.begin() << " leading '0xf8' ";
	bytes.erase(bytes.begin(), start);
    }

#if 0
    // it seems that a few '0xf8' are needed at the end!!!
    if (!bytes.empty())
    {
	// strip following stuff:
	// cannot use reverse_iterator for erase() call :-(
	MidiData::iterator end = bytes.end();
	while (end != bytes.begin())
	{
	    --end;
	    if (STRIPVAL != *end) 
	    {
		++end; // points to first following STRIPVAL

		// keep at least 8 STRIPVALs:
		for (unsigned int i = 0; (i < 8) && (end != bytes.end()); ++i)
		    ++end;
		break;
	    }
	}
	if (end != bytes.end())
	{
	    if (verbose > 1) std::cout << bytes.end() - end << " following '0xf8' ";
	    bytes.erase(end, bytes.end());
	}
    }
#endif

    //------------------------------------
    // rewrite vector of bytes into file:
    //------------------------------------
    write_file(filename + ".stripped", bytes, verbose);
//    write_file(filename, bytes, verbose);

    if (verbose) std::cout << " [done]" << std::endl;
} // end of strip()


// --- global functions ---


int main(int argc, char * const argv[])
{
    // loop through command line options:
    int option;
    int verbose = 0;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'h': usage(); exit(0);
	case 'v': ++verbose; break;
	case 'V': print_version(); exit(0);
	default : usage(); exit(1);
	}
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing file" << endl;
	usage();
	exit(1);
    }

    // loop through the remaining arguments: these are hopefully wav files
    unsigned int errnum = 0;
    for (int i=optind; i < argc; i++)
    {
	std::string filename = argv[i];
	try
	{
	    // create a wav file object and let it print some info
	    strip_file(filename, verbose);
	}
	catch(ErrorClass& error)
	{
	    errnum++;
	    cerr << "Error processing " << filename << ": " << error.getMsg() 
		 << endl;
	}
    }

    exit (errnum ? 1 : 0);
    return  (errnum ? 1 : 0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
