//---------------------------------------------------------------------------
// Copyright 2003 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/mlscorr/main.c,v $
// $Id: main.c,v 1.4 2004/04/04 22:39:06 ah Exp $ 
//
// Description: correlation of files
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: main.c,v 1.4 2004/04/04 22:39:06 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <stdlib.h>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
USING_NAMESPACE(std)

#include <AH/Config/Types.h>
#include <AH/Etech/Types.h>
#include <AH/Etech/Algo.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/IO/Plot.h>
#include <AH/STL/String.h>
#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

#include "mlscorr.h"

// --- local definitions ---

// --- local constants ---

static const char* cmdLineOptions = "c:dhnpr:v";
static const char* version = "0.7";

// --- local data ---

static unsigned int debon = 0;
static bool normalize = false;
static bool showplot = false;
static unsigned int channel = 1;

// --- local classes ---

// --- local functions ---


static void usage()
{
    cerr << "Usage: mlscorr [-" << cmdLineOptions << 
	"] mlstaps inputfile outputfile\n"
	 << "       make an MLS cross correlation.\n"
	 << "  options:\n"
	 << "    h         - print this help message\n"
	 << "    c <num>   - channel number (for multi channel WAV files)\n"
	 << "                stereo: 1=left 2=right (default=1)\n"
	 << "    d         - increase debug output (may be given more than once)\n"
	 << "    n         - normalize to highest point = 1 (not for WAV)\n"
	 << "    p         - show plot on screen\n"
	 << "    v         - print version\n"
	 << "  mlstaps     - tabs of MLS separated by '/', e.g. \"16/5/3/2\"\n"
	 << "  input file  - WAV file or 16 bit mono PCM\n"
	 << "  output file - text file for gnuplot or mono WAV file" << endl;
} // end of usage()



static void print_version()
{
    cerr << "mlscorr Vs " << version << " " 
	 << "(c) 2003 Thomas Ludwig, Patrick Piecha, Andreas Hnnebeck.\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()



// read input data from PCM file into buffer
static unsigned int read_pcm_file(corr_t* data, unsigned int num, FILE* fpi)
{
    if (debon > 1) cerr << "read_file: read " << num << " points" << endl;
    unsigned int i = 0;

    int ret = 0;
    do
    {
	const unsigned int snum = 65536;
	int16 sbuf[snum];

	// set size of data to read on this turn:
	unsigned int rnum = snum;
	if (rnum > num) rnum = num;

	// read input buffer:
	if (debon > 2) cerr << "read_file: read " << rnum << " points" << endl;
	ret = fread(sbuf, sizeof(int16), rnum, fpi);
	if (debon > 2) cerr << "read_file: read returns " << ret << endl;
	if (-1 == ret) throw ErrorSys();
	for (int k = 0; k < ret; ++k)
	    data[i++] = sbuf[k];
	num -= ret;
	if ((unsigned int)ret < rnum) break;
    } while (num > 0);

    if (debon > 1) cerr << "read_file: returns " << i << endl;
    return i; // number of read entries
}


// read input data from wav file into buffer
static uint32 read_wav_file(corr_t* data, uint32 num, IORiffWaveFile& wave)
{
    if (debon > 1) cerr << "read_file: read " << num << " points" << endl;
    uint32 i = 0;


    // get size of data for desired samples
    const uint32 snum = 16384; // number of samples to read
    const uint32 bufsize = wave.getDataBufSize(snum);

    int ret = 0;
    do
    {
	unsigned char sbuf[bufsize];
	memset(sbuf, 0, bufsize);

	// set size of data to read on this turn:
	uint32 rnum = snum;
	uint32 rbufsize = bufsize;
	if (rnum > num)
	{
	    rnum = num;
	    rbufsize = wave.getDataBufSize(num);
	}

	// read input buffer:
	if (debon > 2) cerr << "read_file: read " << rnum << " points" << endl;
	ret = wave.readRaw(sbuf, rbufsize);
	if (ret <= 0) break;
	if ((uint32)ret < rbufsize)
	    cerr << "Warning: could not read " << (rbufsize - ret) 
		 << " bytes from WAV file" << endl;  
	for (uint32 k = 0; k < rnum; ++k)
	{
	    data[i++] = wave.getBufSample(sbuf, k, channel);
	}
	num -= rnum;
    } while (num > 0);

    if (debon > 1) cerr << "read_file: returns " << i << endl;
    return i; // number of read entries
}



bool isWaveFile(const std::string& filename)
{
    const char* fn = filename.c_str();
    int len = strlen(fn);
    if (0 == strcasecmp(".wav", fn + len - 4))
	return true;
    return false;
}



// get normalization factor 
static double getNormFact(uint32 num, const corr_t* data)
{
    if (!normalize) return 1.0;

    // normalize at highest point
    if (debon) cerr << "normalize data at highest point" << endl;
    corr_t max = 0;
    corr_t min = 0;
    for (unsigned int i = 0; i < num; ++i)
    {
	corr_t val = data[i];
	if (val > max) max = val;
	if (val < min) min = val;
    }
    min = -min;
    if (min > max) max = min;
    if (max == 0) return 1.0;

    double norm_fact = 1.0 / max;
    if (debon > 1) 
	cerr << "normalize data : peak = " << max << " norm_fact = "
	     << norm_fact << endl;
    return norm_fact;
}



template <class corr_t> static void calc_stat(
    const std::string&  head, corr_t* data, unsigned int num)
{
    // determine min, mean and max value:
    corr_t mean = 0;
    corr_t max = 0;
    corr_t min = 0;
    for (unsigned int i = 0; i < num; ++i)
    {
	mean += data[i];
	if (data[i] > max) max = data[i];
	if (data[i] < min) min = data[i];
    }
    mean /= num;
    cerr << head << ": min = " << min << ", mean = " << mean 
	 << ", max = " << max << "; num=" << num << endl;
}


// read input data from file into buffer, return buffer
static corr_t* read_file
(
    const string& in,
    uint32& num,      // number of samples
    uint32& sr,       // sample rate in Hz
    uint32& sb        // number of sample bits
)
{
    //------------------------------------
    // read the data from the input file:
    //------------------------------------

    corr_t* din = 0;      // input samples (int32)
    if (isWaveFile(in))
    {
	IORiffWaveFile wave(in.c_str(), false, debon);
	if (debon > 1) wave.printInfo(cerr);
	uint32 chnum = wave.getChannelNum();
	if (channel > chnum) 
	    throw ErrorBasic("illegal argument following option -c");

	num = wave.getSampleNum();
	sb = wave.getSampleSize();
	uint32 sr_in = wave.getSampleRate();

	if (0 == sr) sr = sr_in;
	else if (sr != sr_in) throw ErrorBasic("non matching samplerate");

	// create input buffer
	din = new corr_t[num];
	if (!din) throw ErrorSysOutOfMem();

	// read input data into input buffer:
	if (debon) cerr << "Input: reading wav file " << in << endl;
	uint32 gotten = read_wav_file(din, num, wave);
	if (((int)num - (int)gotten) > 0)
	{
	    cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << endl;
	    num = gotten;
	}
    }
    else // binary PCM file
    {
	// get size of file:
	struct stat st;
	int ret = stat(in.c_str(), &st);
	if (0 != ret) throw ErrorSys("Cannot stat " + in);

	// determine number of data points:
	num = st.st_size / sizeof(short);
	if (debon > 2) cerr << in << ": size=" << st.st_size << " = " << num
			    << " points" << endl;
	if (2 > num) throw ErrorBasic("Size of " + in + " is too small");

	// open data file for read:
	FILE* fpi = fopen(in.c_str(), "rb");
	if (!fpi) throw ErrorSys("Cannot open " + in);

	// create input buffer
	din = new corr_t[num];
	if (!din)
	{
	    fclose(fpi);
	    throw ErrorSysOutOfMem();
	}

	// read input data into input buffer:
	if (debon) cerr << "Input: reading PCM file " << in << endl;
	uint32 gotten = read_pcm_file(din, num, fpi);

	// close file, done:
	fclose(fpi);
	if (((int)num - (int)gotten) > 0)
	{
	    cerr << "WARNING: " << (int)num - (int)gotten 
		 << " missing data points in " << in << endl;
	    num = gotten;
	}

    }
    if (debon) calc_stat (in, din, num);
    return din;
}


// write result of correlation into file
static void plot_file(uint32 sr, corr_t* data, uint32 num)
{
    // get normalisation factor if desired:
    double nf = getNormFact(num, data);

    unsigned int psize = 1024;
    if (psize > num) psize = num;

    // create plot object:
    IOPlot plot;

    // set the default settings:
    std::vector<std::string> plotcmd;
    plotcmd.push_back("set grid");
    plotcmd.push_back("set xlabel \"s\"");
    plotcmd.push_back("set ylabel \"%\"");
    std::string title("MLS cross correlation");
    plotcmd.push_back("set title \"" + title + "\"");
#if 0
    plotcmd.push_back(STLStringPrintf(
	64,"set yrange [%.0f:20]",  mindb-10));
#endif
    plot.exec(plotcmd);


    std::string datacmd("plot \"-\" ");
    if (psize == num)
    {
	// just keep current data number
	const double t_inc = 1.0 / (double)sr;
	plot.exec(STLStringPrintf(
	    64,"set xrange [%f:%f]\n", -0.1 * t_inc * num, 1.1 * t_inc * num));

	datacmd += "using 1:2 title \"mls xcor\" with lines\n";
	double t = 0;
	for (unsigned int i = 0; i < num; ++i, t += t_inc)
	{
	    datacmd += STLStringPrintf(64, "%f %f\n", t, data[i] * nf);
	}
    }
    else
    {
	// reduce data number
	const double t_inc = 1.0 / (double)sr * (double)num / (double)psize;
	plot.exec(STLStringPrintf(64,"set xrange [%f:%f]\n", 
				  -0.1 * t_inc * psize, 1.1 * t_inc * psize));
	datacmd += "using 1:2 title \"mls xcor\" with lines\n";

	const double d_inc = (double)num / (double)psize;
	const double dlim_inc = 0.5 * d_inc;
	double t = 0;
	double d = 0;
	for (unsigned int p = 0; p < psize; ++p, d += d_inc, t += t_inc)
	{
	    // determine index span (float) of this plot point:
	    double di1 = d - dlim_inc;
	    double di2 = d + dlim_inc;

	    // determine first and last index:
	    int bi1 = (int)(di1 + 0.5);
	    int bi2 = (int)(di2 + 0.5);
	    if (bi1 < 0) bi1 = 0;
	    if ((unsigned int)bi2 >= num) bi2 = num -1;

	    // determine proportional parts of first and last index:
	    double pi1; // proportional part of bi1
	    double pi2; // proportional part of bi2
	    if (bi1 == bi2)
	    {
		pi1 = 0.5 * fabs(di2 - di1);
		pi2 = pi1;
	    }
	    else
	    {
		pi1 = (double)bi1 + 0.5 - di1;
		pi2 = fabs(di2 - (double)bi2 + 0.5);
	    }

	    // find mean, max and min value:
	    double max = data[bi1];
	    if (max < data[bi2]) max = data[bi2];
	    double min = data[bi1];
	    if (min > data[bi2]) min = data[bi2];

	    // sum up over all indices:
	    double sum = pi1 * data[bi1] + pi2 * data[bi2];
	    double cnt = pi1 + pi2;
	    for (int i = bi1+1; i < bi2; ++i)
	    {
		sum += data[i];
		if (min > data[i]) min = data[i];
		if (max < data[i]) max = data[i];
		++cnt;
	    }
	    double mean = sum / cnt * nf; // mean value
	    max *= nf;
	    min *= nf;

	    datacmd += STLStringPrintf(64, "%f %f\n", t, mean);
	    datacmd += STLStringPrintf(64, "%f %f\n", t, min);
	    datacmd += STLStringPrintf(64, "%f %f\n", t, max);
	    datacmd += STLStringPrintf(64, "%f %f\n", t, mean);
	}
    }
    datacmd += "e\n";
    plot.exec(datacmd);

    cerr << "hit return to continue" << endl;
    char tmp[80];
    (void)fgets(tmp, 79, stdin);
}


// write result of correlation into file
static void write_file(const string& out, uint32 sr, uint32 sb, 
		       corr_t* data, uint32 num)
{
    if (isWaveFile(out))
    {
	// create WAV file:
	int channels = 1;
	IORiffWaveFile wave(out.c_str(), channels, sb, sr, num);
    
	// get size of data for desired samples
	const uint32 snum = 65536; // number of samples to write
	const uint32 bufsize = wave.getDataBufSize(snum);

	// we MUST normalize the data to fit into 'sb' bits:
	corr_t min = 0;
	corr_t max = 0;
	for (unsigned int k = 0; k < num; ++k)
	{
	    corr_t cur = data[k];
	    if (cur > max) max = cur;
	    if (cur < min) min = cur;
	}
	min = -min;
	if (max < min) max = min;
	double norm = (double)wave.getMaxAmplitude()/(double)max;

	uint32 osize = num;
	uint32 i = 0;
	int ret = 0;
	do
	{
	    unsigned char sbuf[bufsize];
	    memset(sbuf, 0, bufsize);

	    // set size of data to read on this turn:
	    uint32 rnum = snum;
	    uint32 rbufsize = bufsize;
	    if (rnum > osize)
	    {
		rnum = osize;
		rbufsize = wave.getDataBufSize(osize);
	    }

	    // write corr_t data (rounded) into output buffer
	    for (uint32 k = 0; k < rnum; ++k)
	    {
		double db_mean = norm * data[i];
		int32 val = (int32)(db_mean + 0.5);
		if (debon > 3)
		    cerr << "write_file: val = " << val << endl;
		for (int chan = 0; chan < channels; ++chan)
		    wave.setBufSample(sbuf, k, chan+1, val);
		++i;
	    }

	    // write output buffer:
	    if (debon > 2) 
		cerr << "write_file: write " << rnum << " points" << endl;
	    ret = wave.writeRaw(sbuf, rbufsize);
	    if (ret <= 0) break;
	    if ((uint32)ret < rbufsize)
		cerr << "Warning: could not write " << (rbufsize - ret) 
		     << " bytes to WAV file" << endl;  
	    osize -= rnum;
	} while (osize > 0);

	wave.writeHeader();
    }
    else
    {
	// get normalisation factor if desired:
	double nf = getNormFact(num, data);

	// open output file
	FILE* fpo = fopen(out.c_str(), "w");
	if (!fpo) throw ErrorSys("Cannot open " + out);

	// print comment line:
	fprintf(fpo, "# num | xcorr\n");
	// print entries:
	for (unsigned int i = 0; i < num; ++i)
	    fprintf(fpo, "%d\t%f\n", i, data[i] * nf);
	fclose(fpo);
    }
}


// do the data processing
static void process(unsigned int mlsPoly,
		    const string& in, const string& out)
{
    //------------------------------------
    // read the data from the input file:
    //------------------------------------
    uint32 numIn = 0;  // number of samples in 2. file
    uint32 sb = 0;   // number of bits in sample
    uint32 sr = 0;       // sample rate in Hz
    corr_t* data = read_file(in, numIn, sr, sb);

    //-------------------------------------------------------
    // calculate the cross correlation from the input data:
    //-------------------------------------------------------
    if (debon) cerr << "executing cross correlation" << endl;
    if (0 != mlsCorrelate(mlsPoly, data, &numIn, debon))
    {
	cerr << "Error in cross correlation" << endl;
	delete[] data;
	return;
    }
    if (debon) calc_stat ("xcorr", data, numIn);

    //-----------------------------------------
    // write the results into the output file:
    //-----------------------------------------
    if (debon) cerr << "writing file " << out << endl;

    write_file(out, sr, sb, data, numIn);
    if (showplot) plot_file(sr, data, numIn);
    delete[] data;
}


// convert MLS taps into polynomial
static unsigned int taps2poly(const std::string mlstaps)
{
    // parse the bits:
    std::vector<std::string> taps;
    STLStringToken(mlstaps, taps, "/");
    if (0 == taps.size()) throw ErrorBasic(mlstaps + ": illegal MLS taps");

    unsigned int poly = 0;
    for (unsigned int i = 0; i < taps.size(); ++i)
    {
	STLStringDelSurround(taps[i]);
	poly |= (1 << atoi(taps[i].c_str())-1);
    }
    if (debon > 1) cerr << "taps = " << mlstaps << ": poly=" << poly << endl;
    return poly;
}


// --- global functions ---


int main(int argc, char * const argv[])
{
    // loop through command line options:
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'd': ++debon; break;
	case 'c': channel = atoi(optarg); break;
	case 'h': usage(); exit(0);
	case 'n': normalize = true; break;
	case 'p': showplot = true; break;
	case 'v': print_version(); exit(0);
	default : usage(); exit(1);
	}
    }
    if (optind >= argc) 
    {
	cerr << "Error: missing file" << endl;
	usage();
	exit(1);
    }

    // loop through the remaining arguments: these are files
    vector<string> args;
    for (int i=optind; i < argc; i++)
    {
	args.push_back(argv[i]);
    }

    std::string mlstaps;
    std::string in_file_1;
    std::string out_file;
    switch (args.size())
    {
    case 0:
    case 1:
    case 2:  cerr << "Error: missing argument(s)" << endl;
	     usage();
	     exit(1);
	     break;
    case 3:  mlstaps = args[0];
	     in_file_1 = args[1];
	     out_file = args[2];
	     break;
    default: cerr << "Error: too many argument(s)" << endl;
	     usage();
	     exit(1);
    }

    // process the files
    static unsigned int errnum = 0;
    try
    {
	process(taps2poly(mlstaps), in_file_1, out_file); 
    }
    catch (ErrorClass& err)
    {
	++errnum;
	cerr << "Error: " << err.getMsg() << endl;
    }
    catch (exception& err)
    {
	++errnum;
	cerr << "Error: " << err.what() << endl;
    }
    
    exit (errnum ? 1 : 0);
    return  (errnum ? 1 : 0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
