/*
//---------------------------------------------------------------------------
// Copyright (C) 2003 Thomas Ludwig, Patrick Piecha (algorithms)
//           (C) 2003 Andreas Hnnebeck (integration into mlscorr program)
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/mlscorr/mlscorr.c,v $
// $Id: mlscorr.c,v 1.5 2004/04/21 20:45:38 ah Exp $ 
//
// Description: correlation of files
//
//---------------------------------------------------------------------------
*/

static const char rcsid[] = "$Id: mlscorr.c,v 1.5 2004/04/21 20:45:38 ah Exp $ (c) 2003 Thomas Ludwig, Patrick Piecha, Andreas Hnnebeck";

/*** include header files ***/

#ifdef LIBRARY_VERSION /* below needed for library version */
#   include <AH/Config/Platform.h> // always the first include
#   include <AH/STL/String.h>
#   include <AH/Error/Sys.h>
#   include <AH/Error/Basic.h>
    USING_NAMESPACE(AH)
#   include <stdio.h>
#   include <stdlib.h>
#   include "mlscorr.h" // own interface

BEGIN_NAMESPACE(AH)

#endif /* LIBRARY_VERSION */


/*** local definitions ***/

#define MAX_ORDER 31
#define MIN_ORDER 3

#ifndef LIBRARY_VERSION
    typedef int corr_t;
#endif /* !LIBRARY_VERSION */

/*** local function prototypes ***/

static unsigned int reverse_polynomial (unsigned int order,
					unsigned int polynomial);
static unsigned int transform_polynomial (unsigned int order,
					  unsigned int polynomial);
static int perm1_direct_int (unsigned int order, unsigned int polynomial,
			     corr_t* source, size_t source_size,
			     size_t source_len, corr_t* dest,
			     size_t dest_size, size_t* dest_len);
static int fht (corr_t* data, size_t data_size, unsigned int order);
static int perm2_direct_int (unsigned int order, unsigned int polynomial,
			     corr_t* source, size_t source_size, 
			     size_t source_len, corr_t* dest, 
			     size_t dest_size, size_t* dest_len);

/*** global functions ***/

#ifdef LIBRARY_VERSION /* below needed for library version */

int mlsCorrelate(unsigned int mlsPoly, corr_t* input, uint32* inputSize,
		 int debon)
{
    // get order of MLS from polynom:
    unsigned int mlsOrder = 0;
    for (unsigned int i = 0; i < 8 * sizeof(mlsPoly); ++i)
	if ((mlsPoly >> i) & 1)
	    mlsOrder = i + 1;

    unsigned int max_len = (1 << mlsOrder);
    size_t buf2_len = 0;
    corr_t* buf2 = new corr_t[max_len];
    if (!buf2) throw ErrorSysOutOfMem();

    if (0 != perm1_direct_int(mlsOrder, mlsPoly, input, max_len, 
                              *inputSize, buf2, max_len, &buf2_len))
    {
        delete[] buf2;
        return -1;
    }

    if (0 != fht (buf2, max_len, mlsOrder))
    {
        delete[] buf2;
        return -1;
    }

    if (0 != perm2_direct_int(mlsOrder, mlsPoly, buf2, max_len, buf2_len,
			       input, max_len, inputSize))
    {
        delete[] buf2;
        return -1;
    }

    delete[] buf2;
    return 0;
}

#endif /* LIBRARY_VERSION */

/*** local functions ***/

/* Thomas Ludwig 2003-04-18 */
/* LFSR in galois-configuration. Patrick Piecha's Funktion als Vorbild */
inline unsigned int lfsr_galois (unsigned int polynomial, unsigned int state)
{
    return ((state >> 1) ^ ((state & 1) * polynomial));
}


/* Patrick Piecha 2003-03-16, lfsr in fibonacci-configuration */
inline unsigned int fib_feedback (unsigned int x)
{
    x ^= (x >> 16);
    x ^= (x >>  8);
    x ^= (x >>  4);
    x ^= (x >>  2);
    x ^= (x >>  1);

    return x & 1;
}


/* Thomas Ludwig 2003-04-27, lfsr in fibonacci-configuration,
   based on Patrick Piecha's function without 'clocks'        */
inline unsigned int lfsr_fib1 (unsigned int order, unsigned int polynomial,
			       unsigned int state)
{
    return (state >> 1) | (fib_feedback (state & polynomial) << (order - 1));
}


/* Thomas Ludwig 2003-04-30
   reverses the bits of a polynomial, e.g. 1010 0100 -> 0010 0101 */
static unsigned int reverse_polynomial (unsigned int order,
					unsigned int polynomial)
{
    unsigned int i;
    unsigned int reversed_polynomial = 0;

    for (i = 1; i < order; ++i)
    {
	reversed_polynomial |= polynomial & 1;
	reversed_polynomial <<= 1;
	polynomial >>= 1;
    }
    return (reversed_polynomial |= polynomial & 1);
}

/* Thomas Ludwig 2003-05-04
   transforms a polynomial into another polynomial that also creates a MLS. 
   Example: {16,5,3,2}    = 1000 0000 0001 0110 ->
            [16,14,13,11} = 1011 0100 0000 0000                        */ 
unsigned int transform_polynomial (unsigned int order, unsigned int polynomial)
{
    return reverse_polynomial
	(order - 1, polynomial & ((1 << (order - 1)) - 1)) |
	1 << (order - 1);
}



/* Thomas Ludwig 2003-05-03 
   Erste Permutation
   Permutationsvektor wird mit Fibonacci-LFSR erzeugt */
static int perm1_direct_int (unsigned int order, unsigned int polynomial,
			     corr_t* source, size_t source_size,
			     size_t source_len, corr_t* dest,
			     size_t dest_size, size_t* dest_len)
{
    unsigned int i;
    unsigned int state;
    unsigned int len;

    len = (1 << order) - 1;

#ifdef LIBRARY_VERSION
    if (source_len > source_size)
	throw ErrorBasic(STLStringPrintf(80,
	    "perm1_direct_int: source_len=%d > source_size=%d",
	    source_len, source_size));
    if (source_len != len)
	throw ErrorBasic(STLStringPrintf(80,
	    "perm1_direct_int: source_len=%d != len=%d",
	    source_len, len));
    if (dest_size < len + 1)
	throw ErrorBasic("perm1_direct_int: dest_size/len mismatch");
    if (order > MAX_ORDER)
	throw ErrorBasic("perm1_direct_int: order too high");
    if (order < MIN_ORDER)
	throw ErrorBasic("perm1_direct_int: order too low");
#else
    if (source_len > source_size || source_len != len ||
	dest_size < len + 1 || order > MAX_ORDER || order < MIN_ORDER)
	return -1;
#endif

    polynomial = reverse_polynomial (order, polynomial);
    state = 1 << (order -1);
    dest[0] = 0;
    for (i = 0; i < len; ++i)
    {
	dest[state] = source[i];
	state = lfsr_fib1 (order, polynomial, state);	
    }
    *dest_len = 1 << order;

    return 0;
}


/* Thomas Ludwig 2003-04-19
   FHT. Butterfly. Kein Kommentar. */
static int fht (corr_t* data, size_t data_size, unsigned int order)
{
    unsigned int i;
    unsigned int len = (1 << order) - 1;

    if (len > data_size)
#ifdef LIBRARY_VERSION
	throw ErrorBasic("fht: mismatch in size");
#else
	return -1;
#endif

    for (i = 1; i <= len; i <<= 1)
    {
	unsigned int j;
	for (j = 0; j < len; j += (i << 1))
	{
	    unsigned int k;
	    for (k = 0; k < i; ++k)
	    {
		corr_t temp = data [j + k];
		data [j + k] += data [j + k + i];
		data [j + k + i] = temp - data [j + k + i];
	    }
	}
    }
    return 0;
}


/* Thomas Ludwig 2003-05-03 
   Zweite Permutation 
   Permutationsvektor wird mit Galois-LFSR erzeugt */
static int perm2_direct_int (unsigned int order, unsigned int polynomial,
			     corr_t* source, size_t source_size,
			     size_t source_len, corr_t* dest,
			     size_t dest_size, size_t* dest_len)
{
    unsigned int i;
    unsigned int len;
    unsigned int state;

    len = (1 << order) - 1;
    if (source_len > source_size || len > dest_size ||
	len + 1 != source_size || order > MAX_ORDER || order < MIN_ORDER)
#ifdef LIBRARY_VERSION
	throw ErrorBasic("perm2_direct_int: mismatch in size");
#else
	return -1;
#endif

    polynomial = transform_polynomial (order, polynomial);
    state = 1 << (order - 1);
    for (i = 0; i < len; ++i)
    {
	dest[i] = source[state];
	state = lfsr_galois (polynomial, state);	
    }
    *dest_len = i;
    return 0;
}


#ifndef LIBRARY_VERSION /* below needed for standalone version */

#include <stdio.h>
#include <time.h>

#   ifndef TRUE
#       define TRUE 1
#   endif
#   ifndef FALSE
#       define FALSE 0
#   endif

#   define SEED 1

/* Hier kann man sich die Ordnung der MLBS und das Polynom aussuchen*/
/* Aesthetisch einwandfreie Version ;-) */
#   if 0
#       define POLY   0x00e10000
#       define ORDER  24
#   endif

#   if 0
#       define POLY   0x00400010
#       define ORDER  23
#   endif

#   if 1
#       define POLY   0x00200001
#       define ORDER  22
#   endif

#   if 0
#       define POLY   0x00080004
#       define ORDER  20
#   endif

#   if 0
#       define POLY   0x00040013
#       define ORDER  19
#   endif

#   if 0
#       define POLY   0x00004001
#       define ORDER  15
#   endif

#   if 0
#       define POLY   0x00000012
#       define ORDER  5
#   endif

#   if 0
#       define POLY   0x00000005
#       define ORDER  3
#   endif

#   define MAX_LEN  (1 << ORDER)

static void profiler (const char* s);
static int gen_mls (unsigned int order, unsigned int polynomial,
		    unsigned int seed, corr_t* mls, size_t size,
		    size_t* mls_len);
static int test_mls_autocorr (corr_t* data, size_t data_size, 
			      unsigned int order, int* mls_ok,
			      unsigned int* delay);


int main(int argc, char* argv[])
{
	static corr_t buf1[MAX_LEN];
	static corr_t buf2[MAX_LEN];
	size_t buf1_len, buf2_len;
	int error, mls_ok;
	unsigned int delay;
	
	printf ("MLS Ordnung ist %u\n\n", ORDER);
	printf ("Test-MLS erzeugen\n");
	profiler ("gen_mls");
	error = gen_mls (ORDER, POLY, SEED, buf1, MAX_LEN, &buf1_len);
	profiler ("gen_mls");
	if (error)
		printf ("Error in gen_mls.");
	printf ("\n");

	printf ("Permutation 1 direkt ausfuehren\n");
	profiler ("perm1_direct_int");
	error = perm1_direct_int (ORDER, POLY, buf1, MAX_LEN, buf1_len, buf2, MAX_LEN, &buf2_len);
	profiler ("perm1_direct_int");
	if (error)
		printf ("Error in perm1_direct_int.\n");
	printf ("\n");

	printf ("FHT ausfuehren\n");
	profiler ("fht");
	error = fht (buf2, MAX_LEN, ORDER);
	profiler ("fht");
	if (error)
		printf ("Error in fht.\n");
	printf ("\n");

	printf ("Permutation 2 direkt ausfuehren\n");
	profiler ("perm2_direct_int");
	error = perm2_direct_int (ORDER, POLY, buf2, MAX_LEN, buf2_len, buf1, MAX_LEN, &buf1_len);
	profiler ("perm2_direct_int");
	if (error)
		printf ("Error in perm2_direct_int.\n");

	printf ("\n");

	printf ("Ergebnis testen:\n");
	error = test_mls_autocorr (buf1, MAX_LEN, ORDER, &mls_ok, &delay);
	if (error)
		printf ("Error in test_mls_autocorr.\n");
	else if (mls_ok)
		printf ("MLS ok. delay = %u\n", delay);
	else
		printf ("MLS NOT ok.\n");
	return 0;
}


/* pp, 2003-03-14 */
static void profiler (const char* s)
{
	static int      f = 0;
	static clock_t  t = 0;

	if (0 == f)
	{
		f = 0 == f ? 1 : 0;
		printf ("%s: ", NULL == s ? "profiler" : s);
		t = clock();
	}
	else
	{
		t = clock() - t;
		printf ("%ld clocks = %f s\n", t, (double)t / CLOCKS_PER_SEC);
		f = 0 == f ? 1 : 0;
	}
}

/* tl 2003-04-27 */
/* Erzeugt eine MLS. Die MLS dient nur als Testsignal. */
int gen_mls (unsigned int order, unsigned int polynomial, unsigned int seed,
			 corr_t* mls, size_t size, size_t* mls_len)
{
	unsigned int i = 0;
	unsigned int state = seed;
	unsigned int len;

	len = (1 << order) - 1;
	if (len > size)
	{
		*mls_len = 0;
		return -1;
	}

	do
	{
		mls[i] = (state & 1)? -1: 1;
		state = lfsr_galois (polynomial, state);
	} while (++i < len);
	*mls_len = len;
	return 0;
}

/* tl 2003-04-18*/
int test_mls_autocorr (corr_t* data, size_t data_size, unsigned int order,
		       int* mls_ok, unsigned int* delay)
{
	unsigned int i, len = (1 << order) - 1;
	unsigned int peak_count = 0, minus_one_count = 0, other_count = 0;

	if (len > data_size)
		return -1;
	for (i = 0; i < len; i++)
	{
		if (data[i] == -1)
			minus_one_count++;
		else if (data[i] == (int)len)
		{
			*delay = i;	
			peak_count++;
		}
		else
			other_count++;
	}
	*mls_ok = (peak_count == 1 && minus_one_count == len - 1 && other_count == 0) ? TRUE : FALSE;
	return 0;
}

#else /* LIBRARY_VERSION */
END_NAMESPACE
#endif /* LIBRARY_VERSION */
