//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/wavtool/wavavg.c,v $
// $Id: wavavg.c,v 1.5 2004/04/21 21:42:39 ah Exp $ 
//
// Description: cut a pice out of a wav file into a new wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: wavavg.c,v 1.5 2004/04/21 21:42:39 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string>
#include <cerrno>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <string>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/IO/Convert.h>
#include <AH/IO/Riff/WaveFile.h>
USING_NAMESPACE(AH)

// --- local constants ---

static const char* cmdLineOptions = "hS:vV";
static const char* version = "0.3";

// reduced as cygwin has problems with big arrays on the stack
// static const uint32 MAX_SAMPLES = 256 * 1024;  // 256k
static const uint32 MAX_SAMPLES = 64 * 1024;  // 64k

// --- local variables ---

static int verbose = 0;

// --- local classes ---

// --- local functions ---

static void usage(bool help = false)
{
    cerr << "Usage: wavavg [-" << cmdLineOptions << "] navg nsam infile outfile\n"
	 << "       print the data from a WAV file into a text file.\n"
	 << "  options:\n"
	 << "   -h     - print this help message\n"
	 << "   -S     - start of averaging (default = 0 = begin of infile)\n"
	 << "   -v     - verbose mode (increase debug level)\n"
	 << "   -V     - print version\n" 
	 << "  navg    - number of averages\n" 
	 << "  nsam    - number of samples to average\n" 
	 << "  infile  - WAV file for input\n" 
	 << "  outfile - WAV file for output\n" 
	 << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << "  wavavg 16 1000 long.wav ave.wav\n"
	 << "  - read 16 x 1000 samples from long.wav, average it and write the\n"
	 << "    resulting 1000 samples into ave.wav.\n"
	 << "  wavavg -S 500 16 1000 long.wav ave.wav\n"
	 << "  - same as above, but start reading at sample 500.\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "wavavg Vs " << version << " " 
	 << "(c) 2003 Andreas Hnnebeck\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to andreas@huennebeck-online.de\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()


//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static void process
(
    uint32 numAverages,
    uint32 numSamples,
    uint32 startSample,
    const string& inputFileName,
    const string& outputFileName
)
{
    if (2 < verbose) cerr << "process:" << endl;

    // create a wav file object for read:
    if (3 < verbose) cerr << "open input file \"" << inputFileName
			  << "\":" << endl;
    IORiffWaveFile in(inputFileName.c_str(), false, verbose);
    if (verbose  > 3) in.printInfo(cerr);

    // create a wav file object for write:
    if (3 < verbose) cerr << "create output file \"" << outputFileName
			  << "\":" << endl;
    IORiffWaveFile out(outputFileName.c_str(), in.getChannelNum(),
		       in.getSampleSize(), in.getSampleRate(),
		       in.getSampleNum());

    // the buffer used for averaging:
    double accum[in.getChannelNum()][MAX_SAMPLES];

    // get size of data buffer for desired samples
    const uint32 bufSize = in.getDataBufSize(MAX_SAMPLES);
    unsigned char buf[bufSize];

    // read without manipulation
    uint32 sizeToRead = in.getDataBufSize(numSamples) * numAverages;
    in.setDataPos(startSample);             // set read pointer
    
    uint32 curSampleOffset = 0;
    uint32 samplesToRead = numSamples;
    do
    {
	// how much can we read (not more than buffer size) ?
	uint32 curSamplesToRead = samplesToRead;
	if (curSamplesToRead > MAX_SAMPLES) curSamplesToRead = MAX_SAMPLES;
	uint32 curBufSize = in.getDataBufSize(curSamplesToRead);

	uint32 samNumToWrite = 0;
	uint32 sizeToWrite = 0;
	for (uint32 take = 0; take < numAverages; ++take)
	{
	    uint32 curStart = startSample + curSampleOffset + take * numSamples;
	    in.setDataPos(curStart);     // set read pointer

	    // read input file data :
	    int32 rdsize = in.readRaw(buf, curBufSize);
	    if (0 >= rdsize) break;
	    sizeToRead -= rdsize;

	    // process data:
	    // get specified sample from buffer
	    uint32 samNum = MAX_SAMPLES;
	    if ((uint32)rdsize < bufSize)
		samNum = rdsize / in.getDataBufSize((uint32)1);
	    if (0 == take)
	    {
		samNumToWrite = samNum; 
		sizeToWrite = rdsize; 
	    }

	    for (uint32 i = 0; i < samNum; ++i)
	    {
		for (uint32 chan = 0; chan < in.getChannelNum(); ++chan)
		{
		    int32 value = in.getBufSample(buf, i, chan); 
		    if (take) accum[chan][i] += value; // akkumulate
		    else      accum[chan][i] = value; // set
		}
	    }
	}

	// divide each akkumulated value and store it again the buffer
	for (uint32 i = 0; i < samNumToWrite; ++i)
	{
	    for (uint32 chan = 0; chan < in.getChannelNum(); ++chan)
	    {
//		int32 value = (int32)((accum[chan][i] / (double) numAverages) + 0.5);
		int32 value = (int32)accum[chan][i]; // for test only
		out.setBufSample(buf,i, chan, value);
	    }
	}

	// write the buffer into the output file:
	uint32 sizeWritten = 0;
	while (0 < sizeToWrite)
	{
	    if (4 < verbose) cerr << "write " << sizeToWrite 
				  << " bytes into output file \"" 
				  << outputFileName << "\":" << endl;
	    sizeWritten = out.writeRaw(buf, sizeToWrite);
	    if (0 >= sizeWritten) break;
	    sizeToWrite -= sizeWritten;
	}
	if (sizeToWrite)
	    cerr << "Warning: could not write " << sizeToWrite  
		 << " bytes into output file \"" 
		 << outputFileName << "\"" << endl;
    }
    while (sizeToRead > 0);
    out.writeHeader();
} // end of process()


// --- global functions ---


int main(int argc, char* argv[])
{
    int startSample = 0;

    //------------------------------
    // handle command line options:
    //------------------------------
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'S': startSample = atoi(optarg); break;
	case 'v': verbose++;            break;
	case 'h': usage(true);          exit(0);
	case 'V': print_version();      exit(0);
	default : usage();              exit(1);
	}
    }

    // handle the remaining 4 arguments: 
    if (argc < optind + 4) 
    {
	cerr << "Error: missing arguments" << endl;
	usage();
	exit(1);
    }
    int navg = atoi(argv[optind++]);
    int nsam = atoi(argv[optind++]);
    string inputFileName(argv[optind++]);
    string outputFileName(argv[optind++]);

    // check for bad arguments:
    if (navg < 2)
    {
	cerr << "Error: navg is out of range (>1)" << endl;
	usage();
	exit (1);
    }
    if (nsam < 1)
    {
	cerr << "Error: nsam is out of range (>0)" << endl;
	usage();
	exit (1);
    }
    if (startSample < 0)
    {
	cerr << "Error: argument to option -S is out of range (>=0)" << endl;
	usage();
	exit (1);
    }

    // check for too many arguments:
    if (optind < argc) 
    {
	cerr << "Error: too many arguments (last valid is \"" 
	     << argv[optind -1] << "\")" << endl;
	usage();
	exit (1);
    }

    //---------------------
    // process input file:
    //---------------------
    try 
    {
	process((uint32)navg, (uint32)nsam, (uint32)startSample,
		inputFileName, outputFileName); 
    }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
