//---------------------------------------------------------------------------
// Copyright 2000 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/wavtool/wavstat.c,v $
// $Id: wavstat.c,v 1.25 2005/08/01 22:52:34 ah Exp $ 
//
// Description: cut a pice out of a wav file into a new wav file
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: wavstat.c,v 1.25 2005/08/01 22:52:34 ah Exp $ (c) 2000 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h> // always the first include
USING_NAMESPACE(AH)

#include <cstdlib>
#ifdef NO_UNISTD
#   include <getopt.h>
#else
#   include <unistd.h>
#endif
#include <string>
#include <cerrno>
#include <cmath>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
USING_NAMESPACE(std)

#include <AH/Error/Class.h>
#include <AH/Error/Basic.h>
#include <AH/Error/Sys.h>
#include <AH/STL/String.h>
#include <AH/IO/Convert.h>
#include <AH/IO/Riff/WaveFile.h>
#include <AH/IO/Riff/WaveFileCached.h>
#include <AH/IO/Plot.h>
#include <AH/IO/Profile.h>
#include <AH/IO/Smart/Buffer.h>
USING_NAMESPACE(AH)

// --- local definitions ---

#define FAST_ALG 1
#define USE_FK_CPD 0
#define PROFILING 0

// --- local constants ---

#if PROFILING
static const char* cmdLineOptions = "Ccdg:fhpP:st:vV";
#else
static const char* cmdLineOptions = "Ccdg:fhP:st:vV";
#endif
static const char* version = "0.9";

// reduced as cygwin has problems with big arrays on the stack
// static const uint32 MAX_SAMPLES = 256 * 1024;  // 256k
static const uint32 MAX_SAMPLES = 64 * 1024;  // 64

// --- local variables ---

static int verbose = 0;
static bool slowread = false;
#if PROFILING
static int profiling = 0;
#endif

// --- local classes ---

// --- local functions ---

static void usage(bool help = false)
{
    cerr << "Usage: wavstat [-" << cmdLineOptions << "] <file> [<file>, ..]\n"
	 << "       plot statistics of audio file(s).\n"
	 << "  options:\n"
	 << "   -c             - create cumulative probability distribution plot\n"
	 << "   -d             - create probability distribution density plot\n"
	 << "   -C             - clean intermediate files when plot is finished\n"
	 << "   -h             - print this help message\n"
	 << "   -f             - create full size intermediate file for gnuplot\n"
	 << "   -g <setting>   - additional gnuplot settings:\n"
	 << "                    may be given multiply, e.g.\n"
	 << "                      -g \"set grid\" -g \"set size 1,1\"\n"
#if PROFILING
	 << "   -p             - increase profiling level\n"
#endif
	 << "   -P <plotfile>  - plot result into specified file:\n"
	 << "                    the type of file extension determines the\n"
	 << "                    terminal type of gnuplot. A missing file\n"
	 << "                    extension means that the name is taken as\n"
	 << "                    display terminal (e.g. 'x11' for Unix).\n"
	 << "   -t <title>     - set plot title\n"
	 << "   -s             - use slow read method for 16 Bit samplerate\n"
	 << "   -v             - verbose mode (increase debug level)\n"
	 << "   -V             - print version\n"
	 << "   <file>         - audio file, supported file types:\n"
	 << "                     .wav           : PCM-WAV\n"
	 << "                     .mp3           : MP3\n"
	 << "                     .mpc .mp+ .mpp : MPEG+\n"
	 << "                     .ogg           : OGG-Vorbis\n"
	 << "                     .pac           : LPAC\n"
	 << "                  - preprocessed amplitude distribution file,\n"
	 << "                    supported file types:\n"
	 << "                     .stat          : uncompressed (256 kB)\n"
	 << "                     .stat.gz       : compressed with gzip\n"
	 << "                     .stat.bz2      : compressed with bzip2\n"
	 << "\n"
	 << "Notes:\n"
	 << "1. For all audio files except PCM-WAV an appropriate decoder\n"
	 << "   must be exist (mpg123, oggdec, mppdec, lpac; check PATH).\n"
	 << "2. For compressed distribution files an appropriate uncompress\n"
	 << "   program must be exist (gunzip, bunzip2; check PATH).\n"
	 << "3. Preprocessed amplitude distribution files are binary data files\n"
	 << "   (little endian) which contain 65536 32bit unsigned integer values.\n"
	 << "   Each of these values contains the number of samples which have the\n"
	 << "   amplitude corresponding to the values index-32768.\n"
	 << endl;
    if (!help) return;

    cerr << "\n"
	 << "Examples:\n"
	 << "\n"
	 << "  wavstat a.wav b.wav\n"
	 << "    This creates the text files a.sigma and b.sigma, which can\n"
	 << "    be plotted manually later.\n"
	 << "\n"
	 << "  wavstat -P x11 a.wav b.wav\n"
	 << "    This plots the statistical data on a Unix display. The\n"
	 << "    intermediate data files (a.sigma, b.sigma) and the plot\n"
	 << "    command file for gnuplot (x11.gpc) are not removed, so the\n"
	 << "    plot can be recreated later.\n"
	 << "\n"
	 << "  wavstat -P stat.png -C -t \"my statistic:\" a.wav b.wav\n"
	 << "    This plots the statistical data into the graphics file\n"
	 << "    'stat.png'. The title of the plot is 'my statistic'.\n"
	 << "    All intermediate files (a.sigma, b.sigma, stat.gpc) are\n"
	 << "    deleted.\n"
	 << endl;
} // end of usage()


static void print_version()
{
    cerr << "wavstat Vs " << version << "\n" 
	 << "  (c) 2004, 2005 Andreas Hnnebeck\n"
	 << "  (c) 2003 Frank Klemm (algorithms and idea)\n"
	 << "  (c) 2004 Bernd Lngerich (cygwin port)\n"
	 << "Licensed under the terms of the GNU General Public License.\n"
	 << "Report bugs to Andreas Hnnebeck <ah@despammed.com>.\n"
	 << "http://www.huennebeck-online.de"
	 << endl;
} // end of print_version()

#include <limits.h>
#include <float.h>

#if 0
static void
analyseFile ( const char* name )
{
    long     __q   [65536];
    long*    q = __q + 32768;
    long     qmax;
    double   __s   [65536];
    double*  s = __s + 32768;
    long     i;
    double   tot;
    double   sum;
    char     dstname [1024];
    int      len = NULL==strrchr (name, '.')  ?  strlen (name)  :  strrchr (name, '.') - name;

    if ( Stat (name, p) < 0 )
        return;

    sum = 0.;
    for ( i = -32768; i < 32768; i++ ) {
        s[i] = 2. * sum + q[i];
        sum += q[i];
    }

    tot = sum;
    for ( i = 32767; i >= 0; i-- ) {
        s[i] = 2. * sum - q[i];
        sum -= q[i];
    }

    for ( i = -32768; i < 32768; i++ ) {
        if ( s[i-1] != s[i]  ||  s[i] != s[i+1] )
            printf ( "%7.4f\t%6ld\n", inverfc (s[i] / tot), -i );
    }

    sprintf ( dstname, DIRECTORY"/%s.xmgr-2", GetTitleFromFilename (name) );
    if ( NULL == freopen (dstname, "w", stdout) ) {
        stderr_printf ( "Can't create '%s'\n", dstname );
        return;
    }
    qmax = 0;
    for ( i = -32768; i < 32768; i += i != -64 ? 1 : 128 ) {
        if ( q[i] > qmax )
            qmax = q[i];
    }
    printf ( __2, 256.*1.01/sqrt(sum) * (sqrt ( (qmax+1)) - 1 ), len, len, name, len, len, name );
    for ( i = -32768; i < 32768;  i++ ) {
        double tmp = abs(i) < 32 ? (q[32]*(i+32) + q[-32]*(32-i))/64. : q[i];
        printf ( "%6ld\t%8.6f\n", i, 256. / sqrt(sum) * ( sqrt(tmp+1) - 1 ) );
    }

    return;
}
#endif


#if FAST_ALG
static double dierfc (double y)
{
    double  s, t, u, w, x, z;

    // check valid input range:
    if (y >= 2.0) return -8;
    if (y <= 0.0) return 8;

    z = y > 1  ?  2. - y  :  y;
    if (z == 0) return (y > 1 ? -8.:+8.); // wg. log(0)
 
    w = 0.916461398268964 - log (z);
    u = sqrt (w);
    s = (log (u) + 0.488826640273108) / w;
    t = 1 / (u + 0.231729200323405);
    x = u * (1 - s * (s * 0.124610454613712 + 0.5)) -
        ((((-0.0728846765585675 * t + 0.269999308670029) * t +
             0.150689047360223) * t + 0.116065025341614) * t +
             0.499999303439796) * t;
    t = 3.97886080735226 / (x + 3.97886080735226);
    u = t - 0.5;
    s = (((((((((0.00112648096188977922  * u +
                 1.05739299623423047e-4) * u - 0.00351287146129100025) * u -
                 7.71708358954120939e-4) * u + 0.00685649426074558612) * u +
                 0.00339721910367775861) * u - 0.011274916933250487  ) * u -
                 0.0118598117047771104 ) * u + 0.0142961988697898018 ) * u +
                 0.0346494207789099922 ) * u + 0.00220995927012179067;
    s = ((((((((((((s * u - 0.0743424357241784861) * u -
        0.105872177941595488) * u + 0.0147297938331485121) * u +
        0.316847638520135944) * u + 0.713657635868730364 ) * u +
        1.05375024970847138 ) * u + 1.21448730779995237  ) * u +
        1.16374581931560831 ) * u + 0.956464974744799006 ) * u +
        0.686265948274097816) * u + 0.434397492331430115 ) * u +
        0.244044510593190935) * t -
        z * exp (x * x - 0.120782237635245222);
    x += s * (x * s + 1);
    return y > 1  ?   -x  :  +x;
}
#else
static double derfc (double x)
{
    double t, u, y;

    t = 3.97886080735226 / (fabs(x) + 3.97886080735226);
    u = t - 0.5;
    y = ((((((((((0.00127109764952614092) * u + 1.19314022838340944e-4) * u -
                  0.003963850973605135  ) * u - 8.70779635317295828e-4) * u +
                  0.00773672528313526668) * u + 0.00383335126264887303) * u -
                  0.0127223813782122755 ) * u - 0.0133823644533460069 ) * u +
                  0.0161315329733252248 ) * u + 0.0390976845588484035 ) * u +
                  0.00249367200053503304;
    y = ((((((((((((y * u - 0.0838864557023001992) * u -
                            0.119463959964325415 ) * u + 0.0166207924969367356) * u +
                            0.357524274449531043 ) * u + 0.805276408752910567 ) * u +
                            1.18902982909273333  ) * u + 1.37040217682338167  ) * u +
                            1.31314653831023098  ) * u + 1.07925515155856677  ) * u +
                            0.774368199119538609 ) * u + 0.490165080585318424 ) * u +
                            0.275374741597376782 ) * t * exp(-x * x);
    return x < 0  ?  2 - y  :  y;
}
#endif

static double inverfc (double x)
{
#if FAST_ALG // fast: 0.23s with -O6
    return dierfc(x) * sqrt(2.);
#else // slow: 3.6s with -O6
    if (x > 1) return -inverfc(2. - x);

    double ret = 0;
    for (int i = 0; i < 32; ++i)
    {
	double ret2 = ret + 4. / (1LU << i);
        if (derfc(ret2) >= x )
            ret = ret2;
    }
    return ret * sqrt(2.);
#endif
}


//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static uint32* read_preprocessed_values
(
    bool isPipe,
    const std::string& inputCmd,
    uint32& valueSize,
    uint32& idxShift,
    uint32& total,
    int32& minVal,
    int32& maxVal
)
{
    // just in case: defaults are nothing
    total = 0;
    minVal = 0;
    maxVal = 0;
    valueSize = 0;
    idxShift = 0;

    if (2 < verbose) cerr << "read_values:" << endl;

    FILE* fp;
    if (isPipe) fp = popen(inputCmd.c_str(), "r");
    else        fp = fopen(inputCmd.c_str(), "r");
    if (!fp) throw ErrorSys(inputCmd);

    // create array for each possible value:
    int sampleSize = 16;
    valueSize = 1 << sampleSize;
    uint32 maxValue = (1 << (sampleSize - 1)) -1;
    idxShift = valueSize - maxValue - 1; // index starts at 0
    uint32* value = new uint32[valueSize];
    if (!value)
    {
	(void)pclose(fp);
	throw ErrorSys(ENOMEM);
    }

    // initialize array:
    for (uint32 i = 0; i < valueSize; ++i)
	value[i] = 0;

    //----------------------
    // read the whole file:
    //----------------------
    int32 sizeToRead = valueSize * sizeof(value[0]); // bytes to read
    uint8* cbuf = (uint8*)value;
    for (int32 curSize = 0; curSize < sizeToRead; )
    {
	const int32 bufSize = MAX_SAMPLES * 2;

	// read buffer:
	int curBufToRead =  sizeToRead - curSize;
	if (curBufToRead > bufSize) curBufToRead = bufSize;
	int curBufRead = fread(cbuf, sizeof(*cbuf), curBufToRead, fp);
	if (curBufRead <= 0) 
	    break;
	cbuf += curBufRead;
	curSize += curBufRead;
    }
    (void)pclose(fp); fp = 0;

    // determine total value:
    for (uint32 i = 0; i < valueSize; ++i)
    {
	total += value[i];
    }

    // determine min and max values:
    for (uint32 i = 0; i < valueSize; ++i)
    {
	if (value[i] > 0)
	{
	    minVal = i - idxShift;
	    break;
	}
    }
    for (uint32 i = valueSize - 1; i >= 0; --i)
    {
	if (value[i] > 0)
	{
	    maxVal = i - idxShift;
	    break;
	}
    }

    // print finished progress info:
    if (verbose)
	cerr << "reading: " << total/512 
	     << " kB\ncalculating ...";
    if (verbose > 1) cerr << "total=" << total 
			  << ", minVal=" << minVal
			  << ", maxVal=" << maxVal
			  << endl;
    return value;
} // end of read_preprocessed_values()


//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static uint32* read_values
(
    const std::string& inputFileName,
    uint32& valueSize,
    uint32& idxShift,
    uint32& total,
    int32& minVal,
    int32& maxVal,
    std::string& baseName
)
{
    // just in case: defaults are nothing
    total = 0;
    minVal = 0;
    maxVal = 0;
    valueSize = 0;
    idxShift = 0;
    baseName = inputFileName;

    if (2 < verbose) cerr << "read_values:" << endl;

    bool isPipe = false;
    bool preprocessed = false;
    std::string pipe_cmd;

    // determine type of input file:
    std::vector<std::string> token;
    STLStringToken(inputFileName, token, ".");
    if (token.size() > 1)
    {
	int ptnum = token.size() - 1;
	if (token[ptnum-1] == "stat")
	{
	    const std::string devnull = " 2> /dev/null";
	    if (token[ptnum] == "gz")
	    {
		preprocessed = true;
		isPipe = true;
		pipe_cmd = "gunzip -c \"" + inputFileName + "\"" + devnull;
		if (verbose) cerr << "reading compressed preprocessed file '"
				  << inputFileName
				  << "'" << endl;
	    }
	    else if (token[ptnum] == "bz2")
	    {
		preprocessed = true;
		isPipe = true;
		pipe_cmd = "bunzip2 -c \"" + inputFileName + "\"" + devnull;
		if (verbose) cerr << "reading compressed preprocessed file '"
				  << inputFileName
				  << "'" << endl;
	    }
	}
	else if (token[ptnum] == "stat")
	{
	    preprocessed = true;
	    pipe_cmd = inputFileName;
	    if (verbose) cerr << "reading preprocessed file '"
			      << inputFileName
			      << "'" << endl;
	}
    } 

    //--------------------------------
    // open new file for binary read:
    //--------------------------------
    if (preprocessed)
	return read_preprocessed_values(isPipe, pipe_cmd, valueSize, idxShift,
					total, minVal, maxVal);
    
    IORiffWaveFileCached in(inputFileName.c_str(), MAX_SAMPLES, verbose);
    if (verbose  > 3) in.printInfo(cerr);

    // create array for each possible value:
    valueSize = 1 << in.getSampleSize();
    uint32 maxValue = (1 << (in.getSampleSize() - 1)) -1;
    idxShift = valueSize - maxValue - 1; // index starts at 0
    uint32* value = new uint32[valueSize];
    if (!value) throw ErrorSys(ENOMEM);

    // initialize array:
    for (uint32 i = 0; i < valueSize; ++i)
	value[i] = 0;

    //----------------------
    // read the whole file:
    //----------------------
    if (!slowread && in.getSampleSize() == 16)
    {
	//--------------------------------------------------------
	// use fast read method for files with 16 Bit Samplerate:
	//--------------------------------------------------------
	uint32* signedValue = &value[idxShift];

	// how much bytes do we have to read:
	uint32 sizeToRead = in.getSampleNum() * in.getChannelNum() * 2;
	int32 percentStep = sizeToRead / 100;
	for (uint32 curSize = 0; curSize < sizeToRead; )
	{
	    const uint32 bufSize = MAX_SAMPLES * in.getChannelNum() * 2;
	    int16 buf[bufSize];

	    // read buffer:
	    uint32 curBufToRead =  sizeToRead - curSize;
	    if (curBufToRead > bufSize) curBufToRead = bufSize;
	    uint32 curBufRead = in.readRaw((unsigned char*)buf, curBufToRead);
	    if (curBufRead <= 0) 
		break;
	    curSize += curBufRead;

	    curBufRead /= 2;
	    for (uint32 i = 0; i < curBufRead; ++i)
	    {
		// increment array field of matching value
		int32 val = buf[i];
		++signedValue[val];
		if (val < minVal) minVal = val; 
		if (val > maxVal) maxVal = val; 
	    }
	    total += curBufRead;

	    // print progress info:
	    if (verbose)
	    {
		if (in.isPipe()) 
		    cerr << "fast reading: " << total/512/1024 << " MB\r";
		else 
		    cerr << "fast reading: " << curSize/percentStep << " %\r";
	    }
	}

	// print finished progress info:
	if (verbose)
	{
	    if (in.isPipe()) 
		cerr << "fast reading: " << total/512/1024 
		     << " MB\ncalculating ...";
	    else 
		cerr << "fast reading: 100 %\ncalculating ..." << endl;
	}
    }
    else
    {
	//---------------------------------------------------------------
	// use standard read method for files with difficult Samplerate:
	//---------------------------------------------------------------
	uint32* signedValue = &value[idxShift];

	// how many samples do we have to read:
	uint32 samToRead = in.getSampleNum();

	int32 percentStep = samToRead / 100;
	for (uint32 curSam = 0; curSam < samToRead; ++curSam)
	{
	    for (uint32 chan = 0; chan < in.getChannelNum(); ++chan)
	    {
		// read value:
		int32 val;
		if (0 >= in.readSample(&val, curSam, chan))
		    break;

		// increment array field of matching value
		++signedValue[val];
		++total;
		if (val < minVal) minVal = val; 
		if (val > maxVal) maxVal = val; 
	    }

	    // print progress info:
	    if (verbose)
	    {
		if (in.isPipe())
		    cerr << (slowread ? "slow ":"") 
			 << "reading: " << total/512/1024 << " MB\r";
		else if (curSam % percentStep == 0)
		    cerr << (slowread ? "slow ":"") 
			 << "reading: " << curSam / percentStep << " %\r";
	    }
	}

	// print finished progress info:
	if (verbose)
	{
	    if (in.isPipe()) 
		cerr << (slowread ? "slow ":"") << "reading: "
		     << total/512/1024 << " MB\ncalculating ...";
	    else 
		cerr << (slowread ? "slow ":"") 
		     << "reading: 100 %\ncalculating ..." << endl;
	}
    }
    if (verbose > 1) cerr << "total=" << total 
			  << ", minVal=" << minVal
			  << ", maxVal=" << maxVal
			  << endl;
    return value;
} // end of read_values()


static void write_values
(
    const std::string& inputFileName,
    const std::string& outputFileName,
    double* data,
    uint32 size,
    uint32 idxShift,
    double eps,
    const std::string& comments
)
{
    if (verbose) cerr << "writing '" << outputFileName << "'" << endl;

    // create a text file stream for output, print header:
    ofstream out(outputFileName.c_str());
    out << "# Statistics of " << inputFileName << "\n";
    out << comments;
    out << "#--------------------------------------------------------------\n";

    // print out data:
    double lp1_r = 0;   // vorletzt gedrucktes data[lp1_i]
    double lp2_r = 0;   // zuletzt gedrucktes data[lp2_i]

    // set index so that lp2_i - lp_i is not zero at first two points:
    uint32 lp1_i = 0;   // index des vorletzt gedruckten Punkts 
    uint32 lp2_i = 1;   // index des zuletzt gedruckten Punkts

    double m_r = 0;     // Steigung von data[] durch die Punkte
    double b_r = 0;     // offset von data[] durch die Punkte

    for (uint32 valIdx = 0; valIdx < size; ++valIdx)
    {
	int32 val = (int32)valIdx - (int32)idxShift;

	if (eps > 0)
	{
	    //---------------------------------------------------------------
	    // reduce number of data points:
	    // - print only those that are too far away from the value which
	    //   we get if we extrapolate from the last two printed points.
	    //---------------------------------------------------------------
	    bool print = false;
	    if ((valIdx < 2) || (valIdx > (size-3)))
	    {
		// always print the first and last two data points:
		print = true;
	    }
	    else
	    {
		// calculate extrapolated value from two last printed
		// data points for data[]:
		double y = m_r * valIdx + b_r;

		// check if data[] differs enough from y to print it:
		if (fabs(y - data[valIdx]) > eps)
		    print = true; // yes
	    }

	    if (print)
	    {
		if ((valIdx > 2) && (valIdx-1 != lp2_i))
		{
		    // print also previous data point:
		    out << val-1 << "\t" << data[valIdx-1] << "\n";

		    // keep track of printed data points:
		    lp1_r = data[valIdx-1];
		    lp1_i = valIdx-1;
		}
		else
		{
		    // keep track of printed data points:
		    lp1_r = lp2_r;
		    lp1_i = lp2_i;
		}

		// keep track of printed data points:
		lp2_r = data[valIdx];
		lp2_i = valIdx;

		// calculate extrapolation parameters from two last printed
		// data points for data[]:
		m_r = (lp2_r - lp1_r) / (lp2_i - lp1_i);
		b_r = lp2_r - m_r * lp2_i;

		// print data point:
		out << val << "\t" << data[valIdx] << "\n";
	    }
	}
	else
	{
	    // print all data points:
	    out << val << "\t" << data[valIdx] << "\n";
	    continue;
	}
    }
    out << endl;
} // end of write_values


//----------------------------------------------------------
// calculate probability density distribution (AH version):
//----------------------------------------------------------
static double* calc_cpd(const uint32* value, uint32 valueSize, uint32 total)
{
    //---------------------------------------------------
    // use of cpdi prevents sum up of calculation errors
    //---------------------------------------------------
    double* cpd = new double[valueSize];
    if (!cpd) throw ErrorSys(ENOMEM);
    uint32* cpdi = new uint32[valueSize];
    if (!cpdi)
    {
	delete[] cpd;
	throw ErrorSys(ENOMEM);
    }
    cpdi[0] = value[0];
    cpd[0] = (double)cpdi[0] / (double)total;
    for (uint32 i = 1; i < valueSize; ++i) 
    {
	cpdi[i] = cpdi[i-1] + value[i];
	cpd[i] = (double)cpdi[i] / (double)total;
    }
    delete[] cpdi;
    return cpd;
}

//-----------------------------------------------------------
// calculate probability distribution (improved FK version):
//-----------------------------------------------------------
#if USE_FK_CPD
static double* calc_cpd_fk(const uint32* value, uint32 valueSize, uint32 total)
{
    //-----------------------------------------------------------
    // to prevent a sum up of calculation errors we calculate
    // the PDD in two steps:
    //   1. calculate the negative values, starting with 0 
    //      at the lowest index
    //   2. calculate the positive values, starting with 1 
    //      at the lowest index
    // (This still allows a sum up of calculation errors over
    //  the half range.)
    //-----------------------------------------------------------
    double* s = new double[valueSize];
    if (!s) throw ErrorSys(ENOMEM);

    // handle negative values:
    double sum = 0;
    for (uint32 i = 0; i < valueSize/2; ++i) 
    {
        s[i] = 2. * sum + value[i]; // may be "2 * (sum * value)" ?
        sum += value[i];
    }
    // handle positive values:
    sum = (double)total;
    for (uint32 i = valueSize-1; i >= valueSize/2; --i) 
    {
        s[i] = 2. * sum - value[i]; // may be "2 * (sum - value)" ?
        sum -= value[i];
    }
    return s;
}
#endif

//----------------------------------------------------------------
// On Error: throws an exception
//----------------------------------------------------------------
static void process
(
    const std::string& inputFileName,
    bool full,
    std::vector<std::string>& sigma_fn,
    std::vector<std::string>* pdd_fn,
    std::vector<std::string>* cpd_fn
)
{
    // declare pointers here, to be able to delete them after
    // an exception has been thrown:
    uint32* value = 0;
    double* cpd = 0;
#if USE_FK_CPD
    double* s = 0;
#endif

    try
    {
#if PROFILING
	IOProfile prof;
#endif
	if (2 < verbose) cerr << "process:" << endl;

	//--------------
	// read values:
	//--------------
#if PROFILING
	if (profiling > 1) prof.catchNow();
#endif
	int32 minVal = 0;
	int32 maxVal = 0;
	uint32 valueSize = 0;
	uint32 total = 0;
	uint32 idxShift = 0;
	std::string baseName;

	value = read_values(inputFileName, valueSize, idxShift,
			    total, minVal, maxVal, baseName);
#if PROFILING
	if (profiling > 1)
	{
	    cerr << "time for read: " << prof.getLastRunTime() << endl;
	    prof.catchNow();
	}
#endif

	// calculate probability density distribution (AH version):
	cpd = calc_cpd(value, valueSize, total);

#if USE_FK_CPD
	// calculate probability distribution (improved FK version):
	s = calc_cpd_fk(value, valueSize, total);
#endif

#if PROFILING
	if (profiling > 1)
	{
	    cerr << "time for distribution calculation: "
		 << prof.getLastRunTime() << endl;
	    prof.catchNow();
	}
#endif

	// calculate sigma (FK version):
	IOSmartBuffer<double> sigma(valueSize);
#if USE_FK_CPD
	sigma[0] = -inverfc(s[0] / (double)total);
#else
	sigma[0] = -inverfc(2.0 * cpd[0]);
#endif
	double min = sigma[0];
	double max = sigma[0];
	for (uint32 valIdx = 1; valIdx < valueSize; ++valIdx)
	{
#if USE_FK_CPD
	    sigma[valIdx] = -inverfc(s[valIdx] / (double)total);
#else
	    sigma[valIdx] = -inverfc(2.0 * cpd[valIdx]);
#endif
	    if (sigma[valIdx] < min) min = sigma[valIdx];
	    if (sigma[valIdx] > max) max = sigma[valIdx];
	}
#if USE_FK_CPD
	delete[] s; s = 0;
#endif
#if PROFILING
	if (profiling > 1)
	{
	    cerr << "time for sigma calculation: "
		 << prof.getLastRunTime() << endl;
	    prof.catchNow();
	}
#endif

	//----------------------------------
	// create sigma data file for plot:
	//----------------------------------
	std::string outputFileName(baseName + ".sigma");
	std::string comments("# Amplitude value usage (for all channels):\n");
	ostringstream infostream;
	infostream << "# minimum Value = " << minVal << " used " 
		   << value[minVal + idxShift] << " time(s)\n";
	infostream << "# maximum Value = " << maxVal << " used " 
		   << value[maxVal + idxShift] << " time(s)\n";
	infostream.flush();
	comments += infostream.str();
	comments += "# value\tinverfc\n";
	double eps = full ? -1 : (fabs(max - min) / 1.0e4);
	write_values(inputFileName, outputFileName, sigma(), valueSize,
		     idxShift, eps, comments);
	sigma_fn.push_back(outputFileName);

#if PROFILING
	if (profiling > 1)
	    cerr << "time for writing output file" 
		 << (verbose > 2 ? " and sigma calculation":"") 
		 << ": " << prof.getLastRunTime() << endl;
#endif
	if (verbose) cerr << "done: created data file '" 
			  << outputFileName << "'" << endl;
	sigma.remove(); // free used memory, do not use any longer

	if (pdd_fn)
	{
	    //---------------------------------------------
	    // calculate probability distribution density:
	    //---------------------------------------------
	    IOSmartBuffer<double> pdd(valueSize);
	    pdd[0] = (double)value[0] / (double)total;
	    min = pdd[0];
	    max = pdd[0];
	    for (uint32 i = 1; i < valueSize; ++i) 
	    {
		pdd[i] = (double)value[i] / (double)total;
		if (pdd[i] < min) min = pdd[i];
		if (pdd[i] > max) max = pdd[i];
	    }
	    
	    //----------------------------------
	    // create pdd data file for plot:
	    //----------------------------------
	    outputFileName = baseName + ".pdd";
	    comments = "# probability density distribution (for all channels):\n";
	    comments += "# value\tpdd\n";
	    double eps = full ? -1 : (fabs(max - min) / 1.0e4);
	    write_values(inputFileName, outputFileName, pdd(), valueSize,
			 idxShift, eps, comments);
	    pdd_fn->push_back(outputFileName);
	    if (verbose) cerr << "done: created data file '"
			      << outputFileName << "'" << endl;
	}

	if (cpd_fn)
	{
	    //---------------------------------------------------
	    // make cumulative probability distribution density:
	    //---------------------------------------------------
	    // already done above in cpd data.

	    //----------------------------------
	    // create cpd data file for plot:
	    //----------------------------------
	    outputFileName = baseName + ".cpd";
	    comments = "# cumulative probability distribution (for all channels):\n";
	    comments += "# value\tcpd\n";
	    double eps = full ? -1 : 1.0e-4; // max-min = 1
	    write_values(inputFileName, outputFileName, cpd, valueSize,
			 idxShift, eps, comments);
	    cpd_fn->push_back(outputFileName);
	    if (verbose) cerr << "done: created data file '" 
			      << outputFileName << "'" << endl;
	}

	delete[] cpd; cpd = 0;
	delete[] value; value = 0;

#if PROFILING
	if (profiling )
	    cerr << "total run time: " << prof.getRunTime() << endl;
#endif
    }
    catch(const ErrorClass& err)
    {
	cerr << "Error handling '" << inputFileName << "' :" << err.getMsg() 
	     << endl;
    }
    catch(const exception& err)
    {
	cerr << "Error handling '" << inputFileName << "' :" << err.what() 
	     << endl;
    }
    catch(...)
    {
	cerr << "Error handling '" << inputFileName 
	     << "' : unknown exception" << endl;
    }
    if (cpd) delete[] cpd;
    if (value) delete[] value;
#if USE_FK_CPD
    if (s) delete[] s;
#endif
} // end of process()


static void plot_data
(
    const std::string& plotfile,
    const std::string& title,
    const std::string& label,
    const std::vector<std::string>& addSettings,
    std::vector<std::string>& data_fn,
    bool cleanup
)
{
    if (verbose > 1) cerr << "plotfile=\"" << plotfile << "\"" << endl;

    // do some checks:
    if (0 == data_fn.size()) return; // nothing to plot
    if (plotfile == "") return; // nothing to plot

    // determine basename and extension of plotfile:
    std::string baseName;
    std::string terminal;
    bool display;
    string::size_type pos = plotfile.find_last_of(".");
    if (string::npos == pos) 
    {
	// no extension found
	baseName = plotfile;
	terminal = plotfile;
	display = true;
    }
    else
    {
	// separator found:
	baseName = plotfile.substr(0, pos);
	terminal = plotfile.substr(pos+1);
	if (terminal == "") return; // empty terminal: nothing to plot
	display = false;
    }
    if (verbose > 2) cerr << "baseName=\"" << baseName << "\", " 
			  << "terminal=\"" << terminal << "\", " 
			  << "display=" << display << endl;

    std::string usingMode("1:2"); // default
    if (label == "Sigma")
	usingMode = "2:1";

    std::vector<std::string> plotcmd;
    // first the default settings:
    plotcmd.push_back("set grid");
    plotcmd.push_back("set tics out");
    plotcmd.push_back("set key bottom right");
//    plotcmd.push_back("set nokey");
    plotcmd.push_back("set xtics mirror");
    plotcmd.push_back("set ytics mirror");
    plotcmd.push_back("set data style lines");
    if (usingMode == "2:1")
    {
	plotcmd.push_back("set ytics "
			  "(\"32768\" 32768, \"24576\" 24576, "
			  "\"16384\" 16384, \"8192\" 8192, \"0\" 0, "
			  "\"-8192\" -8192, \"-16384\" -16384, "
			  "\"-24576\" -24576, \"-32768\" -32768)");
	plotcmd.push_back("set yrange[-32768:32768]");
	plotcmd.push_back("set xlabel \"" + label + "\"");
	plotcmd.push_back("set ylabel \"Sample\"");
    }
    else
    {
	plotcmd.push_back("set xtics "
			  "(\"32768\" 32768, \"24576\" 24576, "
			  "\"16384\" 16384, \"8192\" 8192, \"0\" 0, "
			  "\"-8192\" -8192, \"-16384\" -16384, "
			  "\"-24576\" -24576, \"-32768\" -32768)");
	plotcmd.push_back("set xrange[-32768:32768]");
	plotcmd.push_back("set xlabel \"Sample\"");
	plotcmd.push_back("set ylabel \"" + label + "\"");
    }
    if (label == "Sigma")
    {
	plotcmd.push_back("set xtics -10, 1, 10");
	plotcmd.push_back("set xrange[-5.5:5.5]");
    }
    else
    {
//	plotcmd.push_back("set ytics 0, 0.1, 1");
//	plotcmd.push_back("set yrange[0:1]");
    }

    if (!display) 
	plotcmd.push_back("set size 1024./128./5., 768./120./4.");
    if (terminal == "png")
	plotcmd.push_back("set terminal png color");
    else
	plotcmd.push_back("set terminal " + terminal);
    plotcmd.push_back("set output \"" + plotfile);
    plotcmd.push_back("set title \"" + title);

    // add additional settings
    for (unsigned int i = 0; i < addSettings.size(); ++i)
    {
        plotcmd.push_back(addSettings[i]);
    }

    // build and add the data plot command:
    std::string cmd("plot ");		     
    for (unsigned int i = 0; i < data_fn.size(); ++i)
    {
	// determine title for each data file:
	std::string title_i;
	string::size_type pos = data_fn[i].find_last_of(".");
	if (string::npos == pos) 
	    title_i = data_fn[i];	// no extension found
	else
	    title_i = data_fn[i].substr(0, pos);

        cmd += (i ? ", \\\n      " : " ");
        cmd += "\"" + data_fn[i];
	cmd += "\" using " + usingMode;
	cmd += " title \"" + title_i + "\"";
    }
    plotcmd.push_back(cmd);
    
    if (!cleanup)
    {
	// determine gnuplot control file: 
	std::string gpcName(baseName + ".gpc");
	if (verbose > 1 ) cerr << "create gnuplot control file '" 
			       << gpcName << "'" << endl;
	ofstream out(gpcName.c_str());
	for (unsigned int i = 0; i < plotcmd.size(); ++i)
	{
	    out << plotcmd[i] << "\n";
	}
	if (display)
	    out << "pause -1\n";
	out.close();
    }

    // do the plot:
    if (verbose) cerr << "plotting '" << plotfile << "'" << endl;
    { 
	IOPlot plot;
	for (unsigned int i = 0; i < plotcmd.size(); ++i)
	{
	    plot.exec(plotcmd[i]);
	}
	if (display)
	{
	    cerr << "hit return to continue" << endl;
	    char tmp[80];
	    (void)fgets(tmp, 79, stdin);
	}
	// make sure that IOPlot has finished before removing
	// the data files (therefore this {} block)
    }

    if (cleanup)
    {
	if (verbose) cerr << "remove intermediate files" << endl;
	for (unsigned int i = 0; i < data_fn.size(); ++i)
	{
	    if (verbose > 1 ) cerr << "remove '" << data_fn[i] << "'" << endl;
	    if (0 != unlink(data_fn[i].c_str()))
		cerr << "cannot delete '" <<  data_fn[i] << "': "
		     << strerror(errno) << endl;
	}
    }
    else
    {
	if (verbose) cerr << "keep intermediate files" << endl;
    }
}


// --- global functions ---


int main(int argc, char* argv[])
{
    // set default options:
    std::string title("statistic");
    std::string plotfile;
    bool full = false;
    bool cleanup = false;
    bool make_cpd = false;
    bool make_pdd = false;
    std::vector<std::string> addSettings;

    //------------------------------
    // handle command line options:
    //------------------------------
    int option;
    while ((option = getopt(argc, argv, cmdLineOptions)) != -1)
    {
	switch (option) 
	{
	case 'c': make_cpd = true;      break;
	case 'd': make_pdd = true;      break;
	case 'C': cleanup = true;       break;
#if PROFILING
	case 'p': ++profiling;          break;
#endif
	case 'f': full = true;          break;
	case 'g': addSettings.push_back(optarg); break;
	case 'P': plotfile = optarg;    break;
	case 's': slowread = true;      break;
	case 't': title = optarg;       break;
	case 'v': ++verbose;            break;
	case 'h': usage(true);          exit(0);
	case 'V': print_version();      exit(0);
	default : usage();              exit(1);
	}
    }

    // the remaining arguments are input files:
    if (optind >= argc) 
    {
	cerr << "Error: missing input file" << endl;
	usage();
	exit(1);
    }
    std::vector<std::string> in_fn;
    while (optind < argc)
	in_fn.push_back(argv[optind++]);

    //---------------------
    // process input file:
    //---------------------
    try 
    { 
	// process all files
	std::vector<std::string> sigma_fn;
	std::vector<std::string> pdd_fn;
	std::vector<std::string> cpd_fn;
	for (unsigned int i = 0; i < in_fn.size(); ++i)
	{
	    // process files
	    process(in_fn[i], full, sigma_fn,
		    (make_pdd ? &pdd_fn : 0),
		    (make_cpd ? &cpd_fn : 0)); 
	}
	plot_data(plotfile, title, "Probability Density",
		  addSettings, pdd_fn, cleanup);
	plot_data(plotfile, title, "Cumulative Probability",
		  addSettings, cpd_fn, cleanup);
	plot_data(plotfile, title, "Sigma", addSettings, sigma_fn, cleanup);
    }
    catch(ErrorClass& error)
    {
	cerr << error.getMsg() << endl;
	exit(1);
    }

    exit (0);
    return  (0); // to keep MSVC quiet
} // end of main()


// --- end of file ---
